# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Trace annotated with debugging information."""

import html
import logging
import os
import re

import voluptuous
import voluptuous.humanize

from cbmc_viewer import markup_link
from cbmc_viewer import srcloct
from cbmc_viewer import templates
from cbmc_viewer import util

# Name of directory holding annotated traces
TRACES = "traces"

################################################################
# Data passed to jinja to generate annotated traces from trace.jinja.html

VALID_STEP = voluptuous.Schema({
    'kind': voluptuous.Any(
        'function-call',
        'function-return',
        'variable-assignment',
        'parameter-assignment',
        'assumption',
        'failure'
    ),
    'num': int,
    'srcloc': str,
    'code': voluptuous.Any(str, None),
    'cbmc': str
}, required=True)

VALID_TRACE = voluptuous.Schema({
    'prop_name': str,
    'prop_desc': str,
    'prop_srcloc': str,
    'steps': [VALID_STEP],
    'outdir': str
}, required=True)

################################################################
# Source code fragments used to annotate traces.

class CodeSnippet:
    """Source code fragments."""

    def __init__(self, root):
        self.root = root  # source root
        self.source = {}  # cache mapping file name -> lines of source code

    def lookup(self, path, line):
        """A line of source code."""

        if line <= 0: # line numbers are 1-based
            logging.info("CodeSnippet lookup: line number not positive: %s", line)
            return None
        line -= 1    # list indices are 0-based

        try:
            if path not in self.source:
                with open(os.path.join(self.root, path), encoding='utf-8') as code:
                    self.source[path] = code.read().splitlines()
        except FileNotFoundError as error:
            if srcloct.is_builtin(path): # <builtin-library-malloc>, etc.
                return None
            raise UserWarning(f"CodeSnippet lookup: file not found: {path}") from error

        # return the whole statement which may be broken over several lines
        snippet = ' '.join(self.source[path][line:line+5])
        snippet = re.sub(r'\s+', ' ', snippet).strip()
        idx = snippet.find(';')     # end of statement
        if idx >= 0:
            return html.escape(snippet[:idx+1])
        idx = snippet.find('}')     # end of block
        if idx >= 0:
            return html.escape(snippet[:idx+1])
        return html.escape(snippet) # statement extends over more that 5 lines


    def lookup_srcloc(self, srcloc):
        """A line of source code (at a source location)."""

        return self.lookup(srcloc['file'], srcloc['line'])

################################################################

class Trace:
    """Trace annotated with debugging information."""

    def __init__(self, name, trace, symbols, properties, loops, snippets,
                 outdir='.'):
        self.prop_name = name
        self.prop_desc = properties.get_description(name)
        self.prop_srcloc = format_srcloc(properties.get_srcloc(name)
                                         or loops.lookup_assertion(name),
                                         symbols)
        self.steps = [{
            'kind': step['kind'],
            'num': num+1, # convert 0-based index to 1-based line number
            'srcloc': format_srcloc(step['location'], symbols),
            'code': snippets.lookup_srcloc(step['location']),
            'cbmc': format_step(step)
        } for num, step in enumerate(trace)]
        self.outdir = outdir
        self.validate()

    def __str__(self):
        """Render annotated trace as html."""

        return templates.render_trace(self.prop_name,
                                      self.prop_desc,
                                      self.prop_srcloc,
                                      self.steps)

    def validate(self):
        """Validate members of an annotated trace object."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_TRACE
        )

    def dump(self, filename=None, outdir=None):
        """Write annotated trace to a file rendered as html."""

        util.dump(self,
                  filename or self.prop_name + ".html",
                  outdir or self.outdir)

################################################################
# Format a source location

def format_srcloc(srcloc, symbols):
    """Format a source location for a trace step."""

    if srcloc is None:
        return 'Function none, File none, Line none'

    fyle, func, line = srcloc['file'], srcloc['function'], srcloc['line']
    func_srcloc = symbols.lookup(func)
    # Warning: next line assumes trace root is subdirectory of code root
    from_file = os.path.join(TRACES, 'foo.html') # any name foo.html will do
    # pylint: disable=consider-using-f-string
    return 'Function {}, File {}, Line {}'.format(
        markup_link.link_text_to_srcloc(func, func_srcloc, from_file),
        markup_link.link_text_to_file(fyle, fyle, from_file),
        markup_link.link_text_to_line(line, fyle, line, from_file)
    )

################################################################
# Format a trace step

def format_step(step):
    """Format a trace step."""

    markup = {
        "function-call": format_function_call,
        "function-return": format_function_return,
        "variable-assignment": format_variable_assignment,
        "parameter-assignment": format_parameter_assignment,
        "assumption": format_assumption,
        "failure": format_failure
    }[step['kind']]
    return markup(step)

def format_function_call(step):
    """Format a function call."""

    name, srcloc = step['detail']['name'], step['detail']['location']

    line = f'-> {markup_link.link_text_to_srcloc(name, srcloc, "./trace/trace.html")}'
    return line

def format_function_return(step):
    """Format a function return."""

    name, srcloc = step['detail']['name'], step['detail']['location']
    line = f'<- {markup_link.link_text_to_srcloc(name, srcloc, "./trace/trace.html")}'
    return line

def format_variable_assignment(step):
    """Format an assignment statement."""

    asn = step['detail']
    lhs, rhs, binary = asn['lhs'], asn['rhs-value'], asn['rhs-binary']
    binary = f'({binary})' if binary else ''
    return f'{lhs} = {rhs} {binary}'

def format_parameter_assignment(step):
    """Format an assignment of an actual to formal function argument."""

    asn = step['detail']
    lhs, rhs, binary = asn['lhs'], asn['rhs-value'], asn['rhs-binary']
    binary = f'({binary})' if binary else ''
    return f'{lhs} = {rhs} {binary}'

def format_assumption(step):
    """Format a proof assumption."""

    pred = step['detail']['predicate']
    return f'assumption: {pred}'

def format_failure(step):
    """Format a proof failure."""

    prop = step['detail']['property'] or "Unnamed"
    reason = step['detail']['reason'] or "Not given"
    return f'failure: {prop}: {reason}'

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""The symbols used to build a goto binary."""

import json
import logging

import voluptuous
import voluptuous.humanize

from cbmc_viewer import ctagst
from cbmc_viewer import parse
from cbmc_viewer import runt
from cbmc_viewer import sourcet
from cbmc_viewer import srcloct
from cbmc_viewer import symbol_table
from cbmc_viewer import util

JSON_TAG = 'viewer-symbol'

################################################################
# Data validator for symbol object

VALID_SYMBOL = voluptuous.Schema({
    'symbols': {
        # symbol name -> symbol srcloc
        voluptuous.Optional(str) : srcloct.VALID_SRCLOC
    }
}, required=True)

################################################################
# Symbol class and subclasses

class Symbol:
    """A mapping from symbols to source locations."""

    def __init__(self, symbols=None):
        """Save and validate a mapping from symbols to source locations."""

        self.symbols = symbols or {}

        # show progress: symbol validation can be slow on large tables
        logging.info('Validating symbol definitions...')
        self.validate()
        logging.info('Validating symbol definitions...done')

    def __repr__(self):
        """A dict representation of a symbol table."""

        # Skip output validation.  It can take 30 seconds on large
        # tables, and the common case is to write the symbol table
        # immediately after building (and validating) the table.
        return self.__dict__

    def __str__(self):
        """A string representation of a symbol table."""

        return json.dumps({JSON_TAG: self.__repr__()},
                          indent=2,
                          sort_keys=True)

    def validate(self):
        """Validate symbols."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_SYMBOL
        )

    def dump(self, filename=None, directory=None):
        """Write symbols to a file or stdout."""

        util.dump(self, filename, directory)

    def lookup(self, symbol):
        """Look up a symbol's source location."""

        return self.symbols.get(symbol)

################################################################

class SymbolFromJson(Symbol):
    """Load symbol table from the output of make-source.

    Given a list of json files containing symbol tables produced by
    make-symbol, merge these symbol tables into a single symbol table.
    """

    def __init__(self, symbol_jsons):

        if not symbol_jsons:
            raise UserWarning('No symbols')

        def load(symbol_json):
            return parse.parse_json_file(symbol_json)[JSON_TAG]

        def merge(tables):
            return util.merge_dicts(tables)

        super().__init__(
            merge([load(symbol_json) for symbol_json in symbol_jsons])
        )

################################################################

class SymbolFromCtags(Symbol):
    """Create a symbol table with ctags."""

    def __init__(self, root, files):
        """Use ctags to list symbols defined in source files under source root."""

        super().__init__(symbols_from_ctags(root, files))

def symbols_from_ctags(root, files):
    """Map symbol names to source locations for symbols defined in files under root."""

    def well_typed_tags(tag):
        """Ensure tag has the correct type"""

        try:
            symbol_, file_, line_ = str(tag['symbol']), str(tag['file']), int(tag['line'])
            assert symbol_ and file_ and line_
            return [{'symbol': symbol_, 'file': file_, 'line': line_}]
        except (AssertionError, ValueError, KeyError):
            logging.info('Skipping tag: "%s"', tag)
            return []

    tags = ctagst.ctags(root, files)
    tags = [tag_ for tag in tags for tag_ in well_typed_tags(tag)]
    tags = sorted(tags, key=lambda tag: (tag['symbol'], tag['file'], tag['line']))

    symbol_map = {}
    for tag in tags:
        symbol = tag['symbol']
        if symbol in symbol_map:
            logging.info('Skipping tag: "%s"', tag)
            continue
        symbol_map[symbol] = srcloct.make_srcloc(tag['file'], None, tag['line'], root, root)
    return symbol_map

################################################################

class SymbolFromGoto(Symbol):
    """Load symbol table from a goto binary.

    Extract symbols listed in the symbol table of a goto binary.  Use
    ctags to extract symbols from source files listed in the symbol
    table (which will find type definitions and preprocessor
    definitions not appearing in the goto symbol table).  Merge these
    lists of symbol definitions, with definitions in the symbol table
    taking precedence over definitions in the ctags output.  This
    still omits definitions from files that list only definitions and
    do not contribute symbols to the symbol table.
    """

    def __init__(self, goto, wkdir, srcdir):

        def symbols_from_table(goto, wkdir, srcdir):
            """Symbols defined in the symbol table"""

            return symbol_table.symbol_definitions(goto, wkdir, srcdir)

        def symbols_from_files(goto, wkdir, srcdir):
            """Symbols defined in files listed in the symbol table"""

            files = symbol_table.source_files(goto, wkdir, srcdir)
            return SymbolFromCtags(srcdir, files).symbols

        table_symbols = symbols_from_table(goto, wkdir, srcdir)
        file_symbols = symbols_from_files(goto, wkdir, srcdir)

        # symbols from symbol table dominate symbols from ctags
        symbols = file_symbols
        symbols.update(table_symbols)

        super().__init__(symbols)

################################################################
# Parse a ctags file
#
# This is not easy.  The default output of ctags identifies the symbol
# definition with a file name and a vi regular expression to locate
# the symbol within the file.  We use the textual output of ctags that
# gives a file name and a line number, but the textual output is
# intended to be human readable and not machine parsable.

EXUBERANT = 'exuberant'
UNIVERSAL = 'universal'
CTAGS = 'ctags'

def have_ctags():
    """Test for existence of exuberant or universal ctags."""

    try:
        ctags_help = runt.run([CTAGS, '--help']).splitlines()[0]
        ctags_tokens = ctags_help.lower().split()
        return (
            (EXUBERANT in ctags_tokens and CTAGS in ctags_tokens) or
            (UNIVERSAL in ctags_tokens and CTAGS in ctags_tokens)
        )
    except FileNotFoundError:
        return False

def run_ctags(root, files, chunk=2000):
    """Run ctags from the given root over the given files.

    Some operating systems limit the number of command line arguments
    (or the length of the command), we we process the list of files in
    chunks.
    """

    if not files:
        return ''

    logging.info('Running ctags on %s files...', len(files))
    ctags_data = []
    while files:
        paths = files[:chunk]
        files = files[chunk:]
        logging.info('Running ctags on %s files starting with %s...',
                     len(paths), paths[0])
        ctags_output = runt.run(
            [CTAGS, '-n', '-f', '-'] + paths, cwd=root, encoding='latin1'
        )
        ctags_data += ctags_output.splitlines()
    logging.info('Running ctags on %s files...done', len(files))
    return ctags_data

def parse_ctags_data(ctags_data, root):
    """Parse the ctags output into a mapping from symbol to source location."""

    logging.info('Parsing ctag data...')

    symbols = {}
    for line in ctags_data:
        # line has the form: symbol<tab>file<tab>line;" ...

        symbol, filename, linenumber = line.split(';"')[0].split("\t")
        srcloc = srcloct.make_srcloc(
            filename, None, linenumber, root, root
        )

        old_srcloc = symbols.get(symbol)
        if old_srcloc:
            logging.debug(
                'Symbol definition: %s: skipping %s, %s; keeping %s %s',
                symbol,
                srcloc['file'], srcloc['line'],
                old_srcloc['file'], old_srcloc['line']
            )
            continue

        symbols[symbol] = srcloc

    logging.info('Parsing ctag data...done')
    return symbols

################################################################
# make-symbol

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_symbol(args):
    """Implementation of make-symbol."""

    viewer_symbol, viewer_source, goto, wkdir, srcdir, files = (
        args.viewer_symbol, args.viewer_source, args.goto, args.wkdir, args.srcdir, None)

    wkdir = srcloct.abspath(wkdir) if wkdir else None
    srcdir = srcloct.abspath(srcdir) if srcdir else None

    # Command line options may enable more than one way
    # to generate the symbol table, so choose the one
    # that gives the most "accurate" results.

    if viewer_symbol:
        logging.info("Symbols by SymbolFromJson")
        return SymbolFromJson(viewer_symbol)

    if viewer_source:
        sources = sourcet.SourceFromJson(viewer_source)
        srcdir = sources.root
        files = sources.files

    if srcdir and files:
        SymbolFromCtags(srcdir, files)

    if goto and wkdir and srcdir:
        logging.info("Symbols by SymbolFromGoto")
        return SymbolFromGoto(goto, wkdir, srcdir)

    logging.info("make-symbol: nothing to do: need "
                 "--goto and --wkdir and --srcdir or "
                 "--viewer-source or"
                 "--viewer-symbol")
    return Symbol()

def make_and_save_symbol(args, path=None):
    """Make symbol object and write to file or stdout"""

    obj = make_symbol(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC coverage checking results.

The Result module describes the results of CBMC property checking, the
Coverage module describes the results of CBMC coverage checking, and
the Property module describes the properties checked during property
checking.
"""

import enum
import json
import logging
import re

import voluptuous
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import srcloct
from cbmc_viewer import util

JSON_TAG = 'viewer-coverage'

################################################################
# Line coverage status

class Status(enum.Enum):
    """Line coverage status."""

    HIT = 1
    MISSED = 2
    BOTH = 3

    def __repr__(self):
        """A string representation of line coverage"""

        # TODO: a JSONEncoder for Status enums so json.dumps(Status.HIT) works

        if self == Status.HIT:
            return "hit"
        if self == Status.MISSED:
            return "missed"
        if self == Status.BOTH:
            return "both"
        raise UserWarning(f"Found unknown line coverage status: {self}")

    def __str__(self):
        """A string representation of line coverage"""

        return self.__repr__()

    @classmethod
    def new(cls, status):
        """Decode a string representation of line coverage"""

        # extend function to the identity function on status instances
        if isinstance(status, Status):
            return status

        if status.lower() == "hit":
            return Status.HIT
        if status.lower() == "missed":
            return Status.MISSED
        if status.lower() == "both":
            return Status.BOTH
        raise UserWarning(f"Found unknown line coverage status: {status}")

    def combine(self, status):
        """Combine line coverage"""

        # extend function to None
        if status is None:
            return self

        # extend function to strings (raise exception for invalid strings)
        status = Status.new(status)

        if self == status:
            return self
        return Status.BOTH

################################################################
# Line coverage validator

LINES_COVERED = voluptuous.Schema(
    {
        'percentage': float, # lines hit / lines total
        'hit': int,
        'total': int
    }, required=True
)

RAW_COVERAGE_DATA = voluptuous.Schema(
    {   # file name ->
        str: {
            # function name ->
            str: {
                # line number -> coverage status
                int: Status
            }
        }
    }, required=True
)

LINE_COVERAGE = voluptuous.Schema(
    {  # file name ->
        str: {
            # line number -> coverage status
            int: Status
        }
    }, required=True
)

FUNCTION_COVERAGE = voluptuous.Schema(
    {  # file name ->
        str: {
            # function name -> (percent, hit, total)
            str: LINES_COVERED
        }
    }, required=True
)

OVERALL_COVERAGE = voluptuous.Schema(LINES_COVERED)


VALID_COVERAGE = voluptuous.Schema(
    {
        'coverage': voluptuous.Any(RAW_COVERAGE_DATA, {}),
        'line_coverage': voluptuous.Any(LINE_COVERAGE, {}),
        'function_coverage': voluptuous.Any(FUNCTION_COVERAGE, {}),
        'overall_coverage': voluptuous.Any(OVERALL_COVERAGE, {})
    }, required=True
)

################################################################
# Line coverage

class Coverage:
    """CBMC coverage checking results"""

    def __init__(self, coverage_list=None):
        """Load CBMC coverage data."""

        coverage_list = coverage_list or []

        # TODO: distinguish between coverage of source code and proof code

        self.coverage = merge_coverage_data(coverage_list)
        self.line_coverage = extract_line_coverage(self.coverage)
        self.function_coverage = extract_function_coverage(self.coverage)
        self.overall_coverage = extract_overall_coverage(self.function_coverage)
        self.validate()

    def __repr__(self):
        """A dict representation of line coverage"""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of line coverage"""

        def serialize(data):
            """Serialize line coverage for the json encoder"""

            if isinstance(data, Status):
                return str(data)
            if isinstance(data, dict):
                return {key: serialize(val) for key, val in data.items()}
            return data

        return json.dumps({JSON_TAG: serialize(self.__repr__())}, indent=2,
                          sort_keys=True)

    def validate(self):
        """Validate coverage."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_COVERAGE
        )

    def dump(self, filename=None, directory=None):
        """Write coverage to a file or stdout."""

        util.dump(self, filename, directory)

    def lookup(self, filename, line):
        """Lookup line coverage status."""

        return self.line_coverage.get(filename, {}).get(line)

def merge_coverage_data(coverage_list):
    """Merge sets of coverage data"""

    coverage = {}
    for coverage_data in coverage_list:
        if not coverage_data:
            logging.info("Found coverage data missing: %s", coverage_data)
            continue

        # file name -> function data
        for filename, function_data in coverage_data.items():

            # The source locations produced by srcloct have paths
            # relative to the source root for all files under the
            # root, and have absolute paths for all other files.
            if filename.startswith('/'):
                logging.info("Restricting coverage to files under the source "
                             "root: skipping %s", filename)
                continue

            coverage[filename] = coverage.get(filename, {})
            # function -> line data
            for func, line_data in function_data.items():
                coverage[filename][func] = coverage[filename].get(func, {})
                # line -> coverage status
                for line, status in line_data.items():
                    cov = coverage[filename][func].get(line)
                    cov = Status.new(status).combine(cov)
                    coverage[filename][func][line] = cov

    try:
        coverage and RAW_COVERAGE_DATA(coverage)
    except voluptuous.Error as error:
        raise UserWarning(f"Error merging coverage data: {error}") from error
    return coverage


# line_coverage: file name -> line number -> status
def extract_line_coverage(coverage):
    """Extract line coverage from raw coverage data."""

    line_coverage = {}
    # file name -> function data
    for filename, function_data in coverage.items():
        line_coverage[filename] = {}
        # function -> line data
        for _, line_data in function_data.items():
            # line -> coverage status
            for line, status in line_data.items():
                line_coverage[filename][line] = status
    return line_coverage

# function_coverage: file name -> function name -> (percentage, hit, total)
def extract_function_coverage(coverage):
    """Extract function coverage from raw coverage data."""

    function_coverage = {}
    # file name -> function data
    for filename, function_data in coverage.items():
        function_coverage[filename] = {}
        # function name -> line data
        for function, line_data in function_data.items():
            hit = 0
            total = 0
            # line number -> coverage status
            for _, status in line_data.items():
                hit = hit + 1 if status != Status.MISSED else hit
                total = total + 1
            percentage = float(hit)/float(total) if total else 0.0
            function_coverage[filename][function] = {
                'percentage': percentage,
                'hit': hit,
                'total': total
            }
    return function_coverage

# overall_coverage: (percentage, hit, total)
def extract_overall_coverage(function_coverage):
    """Extract overall coverage from function coverage data."""

    if not function_coverage:
        return {}

    hit = 0
    total = 0
    # file name -> function data
    for _, function_data in function_coverage.items():
        # function name -> coverage data
        for _, lines_covered in function_data.items():
            hit += lines_covered['hit']
            total += lines_covered['total']
    percentage = float(hit)/float(total) if total else 0.0
    return {
        'percentage': percentage,
        'hit': hit,
        'total': total
    }

################################################################
# Coverage from json output of make-coverage

class CoverageFromJson(Coverage):
    # pylint: disable=too-few-public-methods
    """CBMC coverage results from make-coverage"""

    def __init__(self, json_files):
        super().__init__(
            [load_json(json_file) for json_file in json_files]
        )

def load_json(json_file):
    """Load json file produced by make-coverage"""

    json_data = parse.parse_json_file(json_file)
    if not json_data:
        raise UserWarning(f"Failed to load json coverage data: {json_file}")

    coverage = repair_json(json_data[JSON_TAG]['coverage'])
    try:
        RAW_COVERAGE_DATA(coverage)
    except voluptuous.Error as error:
        raise UserWarning(f"Invalid json coverage data: {json_file}: {error}") from error
    return coverage

def repair_json(coverage):
    """Replace strings with typed values in coverage loaded from json."""

    # Coverage has the type
    #   coverage: file_name -> func_name -> line_num -> HIT/MISSED/BOTH
    # or
    #   proj_data: file -> file_data
    #   file_data: func -> func_data
    #   func_data: line -> status
    # but json represents both line and status as strings

    fix_func = lambda func_data: {int(line_): Status.new(line_data_)
                                  for line_, line_data_ in func_data.items()}
    fix_file = lambda file_data: {func_: fix_func(func_data_)
                                  for func_, func_data_ in file_data.items()}
    fix_proj = lambda proj_data: {file_: fix_file(file_data_)
                                  for file_, file_data_ in proj_data.items()}
    return fix_proj(coverage)

################################################################
# Coverage from json output of cbmc

class CoverageFromCbmcJson(Coverage):
    # pylint: disable=too-few-public-methods
    """CBMC coverage results from cbmc --cover location --json-ui."""

    def __init__(self, json_files, root):
        root = srcloct.abspath(root)
        super().__init__(
            [load_cbmc_json(json_file, root) for json_file in json_files]
        )

def load_cbmc_json(json_file, root):
    """Load json file produced by cbmc --cover location --json-ui."""

    json_data = parse.parse_json_file(json_file)
    if not json_data:
        raise UserWarning(f"Failed to load json coverage data: {json_file}")

    try:
        goals_list = [entry for entry in json_data if 'goals' in entry]
        assert len(goals_list) == 1
        goals = goals_list[0]['goals']
    except AssertionError as error:
        raise UserWarning(
            f"Failed to locate coverage goals in json coverage data: {json_file}") from error

    coverage = {}
    for goal in goals:
        lines = (parse_basicBlockLines(goal.get("basicBlockLines")) or
                 parse_description(goal.get("description")))
        status = goal["status"]
        wkdir = srcloct.json_srcloc_wkdir(goal["sourceLocation"])
        coverage = add_coverage_data(coverage, lines, status, wkdir, root)

    try:
        RAW_COVERAGE_DATA(coverage)
    except voluptuous.Error as error:
        raise UserWarning(f"Invalid json coverage data: {json_file}: {error}") from error
    return coverage

################################################################
# Coverage from xml output of cbmc

class CoverageFromCbmcXml(Coverage):
    # pylint: disable=too-few-public-methods
    """CBMC coverage results from cbmc --cover location --xml-ui."""

    def __init__(self, xml_files, root):
        root = srcloct.abspath(root)
        super().__init__(
            [load_cbmc_xml(xml_file, root) for xml_file in xml_files]
        )

def load_cbmc_xml(xml_file, root):
    """Load xml file produced by cbmc --cover location --xml-ui."""

    xml = parse.parse_xml_file(xml_file)
    if not xml or xml is None:   # Why is 'xml is None' required?
        raise UserWarning(f"Failed to load xml coverage data: {xml_file}")

    coverage = {}
    for goal in xml.iter("goal"):
        lines = (parse_basic_block_lines(goal.find("basic_block_lines")) or
                 parse_description(goal.get("description")))
        status = goal.get("status")
        wkdir = srcloct.xml_srcloc_wkdir(goal.find("location"))
        coverage = add_coverage_data(coverage, lines, status, wkdir, root)

    try:
        RAW_COVERAGE_DATA(coverage)
    except voluptuous.Error as error:
        raise UserWarning(f"Invalid xml coverage data: {xml_file}: {error}") from error
    return coverage

################################################################
# Parse coverage data

def add_coverage_data(coverage, lines, status, wkdir, root):
    """Add to coverage the coverage data reported for a coverage goal"""

    # Warning: What follows assumes that all relative paths appearing
    # in a coverage goal's list of source lines are relative to the
    # same working directory (the working directory in the coverage
    # goal's source location).  This is probably true.  A project can
    # avoid this issue altogether, however, by invoking goto-cc on
    # absolute paths, ensuring that coverage descriptions contain only
    # absolute paths and no relative paths.

    hit = parse_coverage_status(status)
    lines = relative_locations(lines, wkdir, root)

    for path, func, line in lines:
        coverage = update_coverage(coverage, path, func, line, hit)
    return coverage

def relative_locations(locations, wkdir, root):
    """Replace locations with locations relative to the source root."""

    # Warning: Some goto program transformations currently use a
    # documentation string as a file name in source locations.  The
    # result is a source location that has no valid relative form.  We
    # ignore them when generating coverage data.

    def relative_location(loc_file, loc_func, loc_line):
        """The source location relative to the source root."""
        try:
            return (
                srcloct.make_relative_path(loc_file, root, wkdir),
                loc_func,
                loc_line
            )
        except AssertionError: # raised by make_relative_path
            logging.debug(
                "Ignoring an invalid source location in coverage data:"
                " {file: %s, function: %s, line: %s}",
                loc_file, loc_func, loc_line
            )
            return None

    locations = [relative_location(*loc) for loc in locations]
    locations = [loc for loc in locations if loc is not None]
    return locations

def update_coverage(coverage, path, func, line, status):
    """Add to coverage the coverage status of a single line"""

    coverage[path] = coverage.get(path, {})
    coverage[path][func] = coverage[path].get(func, {})
    coverage[path][func][line] = coverage[path][func].get(line, status)
    coverage[path][func][line] = coverage[path][func][line].combine(status)
    return coverage

################################################################
# Extract the source lines contributing to a coverage goal's basic block
#
# Each coverage goal is a basic block in a goto program. Each
# instruction in a basic block has a source location identifying the
# source line that generated the instruction.
#
# CBMC encodes these source lines in the text of the goal description.
# Later versions of CBMC also encode these source lines a xml or json
# subelements of the goal in CBMC's xml or json coverage output.

def parse_lines(string):
    """Extract line numbers from the string encoding of line numbers used in coverage output"""

    # string is an encoding of a set of line numbers like
    #   "1,3,6" -> {1, 3, 6}
    #   "1-3,6,20-32" -> {1, 2, 3, 6, 20, 21, 22}

    lines = set()
    for group in string.split(','):
        bounds = group.split('-')
        if len(bounds) == 1:
            lines.add(int(bounds[0]))
        elif len(bounds) == 2:
            lines.update(range(int(bounds[0]), int(bounds[1])+1))
        else:
            raise UserWarning(f"Unexpected encoding of line numbers: {string}")
    return sorted(lines)

def parse_description(description):
    """Extract basic block source lines from a coverage goal's textual description"""

    try:
        # description is "block N (lines BASIC_BLOCK)"
        basic_block = re.match(r'block [0-9]+ \(lines (.*)\)', description).group(1)

        # basic_block is
        #   chunk1;chunk2;chunk3
        # each chunk is
        #   test.c:foo:3,6-10
        # each function name foo from rust may include embedded semicolons like
        #   main::foo
        lines = [(fyle, func, line)
                 for chunk in basic_block.split(';')
                 for fyle, func_lines in [chunk.split(':', 1)]  # fyle:func:lines -> fyle,func:lines
                 for func, lines in [func_lines.rsplit(':', 1)] # func:lines -> func,lines
                 for line in parse_lines(lines)]
        assert all(fyle and func and line for fyle, func, line in lines)
        return lines
    except (AttributeError, ValueError, AssertionError) as error:
        # AttributeError after match(): 'NoneType' object has no attribute 'group'
        # ValueError after after split()/rsplit(): not enough values to unpack
        # ValueError in parse_lines(): invalid literal for int()
        raise UserWarning(f'Unexpected coverage goal description: "{description}"') from error

def parse_basic_block_lines(basic_block_lines):
    """Extract basic block source lines from a coverage goal's xml subelement"""

    if basic_block_lines is None:
        return []

    try:
        # basic_block_lines is xml
        #   <basic_block_lines>
        #     <line file="test.c" function="foo">3,6-10</line>
        #     ...
        #   </basic_block_lines>
        lines = [(fyle, func, line)
                 for bbl in basic_block_lines.iter("line")
                 for fyle, func, lines in [(bbl.get("file"), bbl.get("function"), bbl.text)]
                 for line in parse_lines(lines)]
        assert all(fyle and func and line for fyle, func, line in lines)
        return lines
    except (ValueError, AssertionError) as error:
        # ValueError in parse_lines(): invalid literal for int()
        raise UserWarning(f'Unexpected coverage goal xml data: "{basic_block_lines}"') from error

def parse_basicBlockLines(basicBlockLines): # pylint: disable=invalid-name
    """Extract basic block source lines from a coverage goal's json data"""

    if basicBlockLines is None:
        return []

    try:
        # basicBlockLines is json
        #   "basicBlockLines": {
        #     "test.c": {
        #       "main": "16,17"
        #     }
        #   }
        lines = [(fyle, func, line)
                 for fyle, fyle_data in basicBlockLines.items()
                 for func, lines in fyle_data.items()
                 for line in parse_lines(lines)]
        assert all(fyle and func and line for fyle, func, line in lines)
        return lines
    except (ValueError, AssertionError) as error:
        # ValueError in parse_lines(): invalid literal for int()
        raise UserWarning(f'Unexpected coverage goal json data: "{basicBlockLines}"') from error

################################################################
# Parse the hit/miss status in a coverage goal

def parse_coverage_status(status):
    """Parse a coverage status"""

    return Status.HIT if status.upper() == 'SATISFIED' else Status.MISSED

################################################################
# make-coverage

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_coverage(args):
    """The implementation of make-coverage."""

    viewer_coverage, srcdir, cbmc_coverage = args.viewer_coverage, args.srcdir, args.coverage

    if viewer_coverage:
        if filet.all_json_files(viewer_coverage):
            return CoverageFromJson(viewer_coverage)
        fail(f"Expected json files: {viewer_coverage}")

    if cbmc_coverage and srcdir:
        if filet.all_json_files(cbmc_coverage):
            return CoverageFromCbmcJson(cbmc_coverage, srcdir)
        if filet.all_xml_files(cbmc_coverage):
            return CoverageFromCbmcXml(cbmc_coverage, srcdir)
        fail(f"Expected json files or xml files, not both: {cbmc_coverage}")

    logging.info("make-coverage: nothing to do: need "
                 "cbmc coverage checking results and --srcdir or "
                 "--viewer-coverage")
    return Coverage()

def make_and_save_coverage(args, path=None):
    """Make coverage object and write to file or stdout"""

    obj = make_coverage(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Parsing of xml and json files with limited error handling."""

from pathlib import Path
from xml.etree import ElementTree

import json
import logging

def parse_xml_file(xfile):
    """Parse an xml file."""

    try:
        return ElementTree.parse(xfile)
    except (IOError, ElementTree.ParseError) as err:
        logging.debug("%s", err)
        raise UserWarning(f"Can't load xml file '{xfile}'") from None

def parse_xml_string(xstr):
    """Parse an xml string."""

    try:
        return ElementTree.fromstring(xstr)
    except ElementTree.ParseError as err:
        logging.debug("%s", err)
        raise UserWarning(f"Can't parse xml string '{xstr[:40]}...'") from None

def parse_json_file(jfile):
    """Parse an json file."""

    try:
        with open(jfile, encoding='utf-8') as data:
            return json.load(data)
    except (IOError, json.JSONDecodeError) as err:
        logging.debug("%s", err)
        raise UserWarning(f"Can't load json file '{jfile}' in {Path.cwd()}") from None

def parse_json_string(jstr):
    """Parse a json string."""

    try:
        return json.loads(jstr)
    except json.JSONDecodeError as err:
        logging.debug("%s", err)
        raise UserWarning(f"Can't parse json string '{jstr[:40]}...'") from None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC traces."""

import json
import logging
import re

import voluptuous
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import srcloct
from cbmc_viewer import util

JSON_TAG = 'viewer-trace'

################################################################
# Every trace step has a KIND and a LOCATION and other DETAIL

VALID_FUNCTION_CALL = voluptuous.Schema({
    'kind': 'function-call',
    'location': srcloct.VALID_SRCLOC, # function call
    'hidden': bool,
    'detail' : {
        'name': str,
        'name-path': voluptuous.Any(str, None),
        'location': srcloct.VALID_SRCLOC # function being called
    }
}, required=True)

VALID_FUNCTION_RETURN = voluptuous.Schema({
    'kind': 'function-return',
    'location': srcloct.VALID_SRCLOC, # function return
    'hidden': bool,
    'detail' : {
        'name': str,
        'name-path': voluptuous.Any(str, None),
        'location': srcloct.VALID_SRCLOC # function being returned to
    }
}, required=True)

VALID_VARIABLE_ASSIGNMENT = voluptuous.Schema({
    'kind': 'variable-assignment',
    'location': srcloct.VALID_SRCLOC,
    'hidden': bool,
    'detail': {
        'lhs': str,
        'lhs-lexical-scope': voluptuous.Any(str, None),
        'rhs-value': str,
        'rhs-binary': voluptuous.Any(str, None)
    }
}, required=True)

VALID_PARAMETER_ASSIGNMENT = voluptuous.Schema({
    'kind': 'parameter-assignment',
    'location': srcloct.VALID_SRCLOC,
    'hidden': bool,
    'detail': {
        'lhs': str,
        'lhs-lexical-scope': voluptuous.Any(str, None),
        'rhs-value': str,
        'rhs-binary': voluptuous.Any(str, None)
    }
}, required=True)

VALID_FAILURE = voluptuous.Schema({
    'kind': 'failure',
    'location': srcloct.VALID_SRCLOC,
    'hidden': bool,
    'detail': {
        'property': voluptuous.Any(str, None),
        'reason': str
    }
}, required=True)

VALID_ASSUMPTION = voluptuous.Schema({
    'kind': 'assumption',
    'location': srcloct.VALID_SRCLOC,
    'hidden': bool,
    'detail' : {
        "predicate": str
    }
}, required=True)

VALID_STEP = voluptuous.Schema(
    voluptuous.Any(
        VALID_FUNCTION_CALL,
        VALID_FUNCTION_RETURN,
        VALID_VARIABLE_ASSIGNMENT,
        VALID_PARAMETER_ASSIGNMENT,
        VALID_FAILURE,
        VALID_ASSUMPTION
    ), required=True
)

VALID_TRACE = voluptuous.Schema(
    [VALID_STEP],
    required=True
)

VALID_TRACES = voluptuous.Schema({
    'traces': {
        # failed property name -> failure trace
        voluptuous.Optional(str): VALID_TRACE}
}, required=True)

################################################################

class Trace:
    """CBMC error traces.

    Given a list of sets of traces (dict representations of the Trace
    object) merge the traces into a single set of results.
    """

    def __init__(self, trace_lists=None):
        trace_lists = trace_lists or []
        self.traces = self.merge(trace_lists)
        self.validate()

    def __repr__(self):
        """A dict representation of traces."""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of traces."""

        return json.dumps({JSON_TAG: self.__repr__()}, indent=2, sort_keys=True)

    def validate(self):
        """Validate tracess."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_TRACES
        )

    def dump(self, filename=None, directory=None):
        """Write traces to a file or stdout."""

        util.dump(self, filename, directory)

    @staticmethod
    def merge(trace_lists):
        """Merge trace lists."""

        def handle_duplicate(key, val1, val2):
            """Handle different traces for the same assertion."""
            _, _ = val1, val2
            logging.warning("Found duplicate traces for property %s", key)

        traces = util.merge_dicts(trace_lists, handle_duplicate)
        traces = {name: close_function_stack_frames(trace)
                  for name, trace in traces.items()}
        traces = {name: strip_external_srclocs(trace)
                  for name, trace in traces.items()}
        return traces

################################################################

class TraceFromJson(Trace):
    """Load error traces from output of make-trace."""

    def __init__(self, json_files):
        super().__init__(
            [load_traces(json_file) for json_file in json_files]
        )

def load_traces(loadfile):
    """Load a trace file."""
    with open(loadfile, encoding='utf-8') as data:
        return json.load(data)[JSON_TAG]

################################################################

class TraceFromCbmcText(Trace):
    """Load error traces from text output of property checking."""

    def __init__(self, text_files, root, wkdir):
        root = srcloct.abspath(root)
        super().__init__(
            [parse_text_traces(text_file, root, wkdir)
             for text_file in text_files]
        )

def parse_text_traces(textfile, root=None, wkdir=None):
    """Parse a set of text traces."""

    with open(textfile, encoding='utf-8') as data:
        lines = '\n'.join(data.read().splitlines())
        blocks = re.split(r'\n\n+', lines)

    traces = {}

    name = None
    trace = []
    in_trace = False
    for block in blocks:
        if block.startswith('Trace for'):
            if name:
                traces[name] = trace
                name = None
                trace = []
            name = block.split()[-1][:-1]
            in_trace = True
            continue
        if not in_trace:
            continue
        if block.startswith('State'):
            trace.append(parse_text_state(block, root, wkdir))
            continue
        if block.startswith('Assumption'):
            trace.append(parse_text_assumption(block, root, wkdir))
            continue
        if block.startswith('Violated property'):
            trace.append(parse_text_failure(block, root, wkdir))
            continue
        if block.startswith('** '):
            if name:
                traces[name] = list(visible_steps(trace))
            break
        raise UserWarning(f"Unknown block: {block}")

    return traces

def parse_text_assignment(string):
    """Parse an assignment in a text trace."""

    # trailing binary expression (exp) may be integer, struct, or unknown ?
    match = re.match(r'([^=]+)=(.+) \(([?{},01 ]+)\)', string.strip())
    if match:
        return list(match.groups()[:3])
    match = re.match('([^=]+)=(.+)', string.strip())
    if match:
        return list(match.groups()[:2]) + [None]
    raise UserWarning(f"Can't parse assignment: {string}")

def parse_text_state(block, root=None, wkdir=None):
    """Parse the state block in a text trace."""

    lines = block.splitlines()
    srcloc = srcloct.text_srcloc(lines[0], wkdir, root)
    # assignment may be split over remaining lines in block
    lhs, rhs_value, rhs_binary = parse_text_assignment(' '.join(lines[2:]))
    return {
        'kind': 'variable-assignment',
        'location': srcloc,
        'hidden': False,
        'detail': {
            'lhs': lhs,
            'lhs-lexical-scope': None,
            'rhs-value': rhs_value,
            'rhs-binary': rhs_binary
        }
    }

def parse_text_assumption(block, root=None, wkdir=None):
    """Parse an assumption in a text trace."""

    # A text assumption block normally has the form
    #   Assumption:
    #     file test.c line 4 function main
    #     x > 0
    # The source location is sometimes missing, but the predicate
    # being assumed is always the last line of the block

    lines = block.splitlines()
    srcloc = srcloct.text_srcloc(lines[1], wkdir, root)
    return {
        'kind': 'assumption',
        'location': srcloc,
        'hidden': False,
        'detail': {
            'predicate': lines[-1].strip()
        }
    }

def parse_text_failure(block, root=None, wkdir=None):
    """Parse a failure in a text trace."""

    # A text failure block normally has the form
    #   Violated property:
    #     file test.c function main line 5 thread 0
    #     assertion x == 0
    #     x == 0
    # The source location is sometimes missing, and the name of the
    # property that failed is always missing in text output, but the
    # reason for the failure is always the next-to-last line in the
    # block

    lines = block.splitlines()
    srcloc = srcloct.text_srcloc(lines[1], wkdir, root)
    return {
        'kind': 'failure',
        'location': srcloc,
        'hidden': False,
        'detail': {
            'property': None,
            'reason': lines[-2].strip()
        }
    }

################################################################

class TraceFromCbmcXml(Trace):
    """Load error traces from xml output of property checking."""

    def __init__(self, xml_files, root):
        root = srcloct.abspath(root)
        super().__init__(
            [parse_xml_traces(xml_file, root) for xml_file in xml_files]
        )

def parse_xml_traces(xmlfile, root=None):
    """Parse a set of xml traces."""

    xml = parse.parse_xml_file(xmlfile)
    if xml is None:
        return {}


    traces = {}

    # cbmc produced all traces as usual
    if xml.find('result') is not None:
        for line in xml.iter('result'):
            name, status = line.get('property'), line.get('status')
            if status == 'SUCCESS':
                continue
            traces[name] = parse_xml_trace(line.find('goto_trace'), root)
        return traces

    # cbmc produced only a one trace after being run with --stop-on-fail
    goto_trace = xml.find('goto_trace')
    if goto_trace is not None:
        failure = goto_trace.find('failure')
        name = failure.get('property') if failure else 'Unknown property'
        traces[name] = parse_xml_trace(goto_trace, root)
        return traces

    # cbmc produced no traces
    return traces

def parse_xml_trace(steps, root=None):
    """Parse a single xml trace."""

    trace = [parse_xml_step(step, root) for step in steps]
    return list(visible_steps(trace))

def parse_xml_step(step, root=None):
    """Parse a step in an xml trace."""

    kind = step.tag
    parser = (parse_xml_failure if kind == 'failure' else
              parse_xml_assignment if kind == 'assignment' else
              parse_xml_function_call if kind == 'function_call' else
              parse_xml_function_return if kind == 'function_return' else
              parse_xml_location_only if kind == 'location-only' else None)

    if parser is None:
        # skip uninteresting kinds of steps
        if kind == 'loop-head':
            logging.debug('Skipping step type: %s', kind)
            return None

        # warn about skipping a potentially interesting kind of step
        logging.warning('Skipping step type: %s', kind)
        return None

    parsed_step = parser(step, root)
    if parsed_step:
        parsed_step['hidden'] = step.get('hidden') == 'true'
    return parsed_step

def parse_xml_failure(step, root=None):
    """Parse a failure step in an xml trace."""

    return {
        'kind': 'failure',
        'location': srcloct.xml_srcloc(
            step.find('location'), root
        ),
        'detail': {
            'property': step.get('property'),
            'reason': step.get('reason')
        }
    }

def parse_xml_assignment(step, root=None):
    """Parse an assignment step in an xml trace."""

    akind = step.get('assignment_type')
    kind = ('variable-assignment' if akind == 'state' else
            'parameter-assignment' if akind == 'actual_parameter' else None)
    if kind is None:
        raise UserWarning(f"Unknown xml assignment type: {akind}")

    return {
        'kind': kind,
        'location': srcloct.xml_srcloc(
            step.find('location'), root
        ),
        'detail': {
            'lhs': step.find('full_lhs').text,
            'lhs-lexical-scope': step.get('identifier'),
            'rhs-value': step.find('full_lhs_value').text,
            'rhs-binary': binary_as_bytes(step.find('full_lhs_value').get('binary'))
        }
    }

def parse_xml_function_call(step, root=None):
    """Parse a function call step in an xml trace."""

    return {
        'kind': 'function-call',
        'location': srcloct.xml_srcloc(
            step.find('location'), root
        ),
        'detail': {
            'name': step.find('function').get('display_name'),
            'name-path': step.find('function').get('identifier'),
            'location': srcloct.xml_srcloc(
                step.find('function').find('location'), root
            )
        }
    }

def parse_xml_function_return(step, root=None):
    """Parse a function return step in an xml trace."""

    return {
        'kind': 'function-return',
        'location': srcloct.xml_srcloc(
            step.find('location'), root
        ),
        'detail': {
            'name': step.find('function').get('display_name'),
            'name-path': step.find('function').get('identifier'),
            'location': srcloct.xml_srcloc(
                step.find('function').find('location'), root
            )
        }
    }

def parse_xml_location_only(step, root=None):
    """Parse (ignore) a location-only step in an xml trace."""

    _ = step
    _ = root

################################################################

class TraceFromCbmcJson(Trace):
    """Load error traces from json output of property checking."""

    def __init__(self, json_files, root):
        root = srcloct.abspath(root)
        super().__init__(
            [parse_json_traces(json_file, root) for json_file in json_files]
        )

def parse_json_traces(jsonfile, root=None):
    """Parse a set of json traces."""

    data = parse.parse_json_file(jsonfile)
    if data is None:
        return {}

    results = [entry['result'] for entry in data if 'result' in entry][0]
    traces = {result['property']: parse_json_trace(result['trace'], root)
              for result in results if 'trace' in result}
    return traces

def parse_json_trace(steps, root=None):
    """Parse a single of json trace."""

    trace = [parse_json_step(step, root) for step in steps]
    return list(visible_steps(trace))

def parse_json_step(step, root=None):
    """Parse a step of a json trace."""

    kind = step['stepType']
    parser = (parse_json_failure if kind == 'failure' else
              parse_json_assignment if kind == 'assignment' else
              parse_json_function_call if kind == 'function-call' else
              parse_json_function_return if kind == 'function-return' else
              parse_json_location_only if kind == 'location-only' else None)
    if parser is None:
        # skip uninteresting kinds of steps
        if kind == 'loop-head':
            logging.debug('Skipping step type: %s', kind)
            return None

        # warn about skipping a potentially interesting kind of step
        logging.warning('Skipping step type: %s', kind)
        return None

    parsed_step = parser(step, root)
    if parsed_step:
        parsed_step['hidden'] = bool(step['hidden'])
    return parsed_step

def parse_json_failure(step, root=None):
    """Parse a failure step of a json trace."""

    return {
        'kind': 'failure',
        'location': srcloct.json_srcloc(
            step.get('sourceLocation'), root
        ),
        'detail': {
            'property': step.get('property'),
            'reason': step.get('reason')
        }
    }

def parse_json_assignment(step, root=None):
    """Parse an assignment step of a json trace."""

    akind = step.get('assignmentType')
    kind = ('variable-assignment' if akind == 'variable' else
            'parameter-assignment' if akind == 'actual-parameter' else None)
    if kind is None:
        raise UserWarning(f"Unknown json assignment type: {akind}")

    # &v is represented as {name: pointer, data: v}
    # NULL is represented as {name: pointer, data: {((basetype *)NULL)}}
    data = step['value'].get('data')
    if step['value'].get('name') == 'pointer' and data and 'NULL' not in data:
        data = f'&{data}'

    return {
        'kind': kind,
        'location': srcloct.json_srcloc(
            step.get('sourceLocation'), root
        ),
        'detail': {
            'lhs': step['lhs'],
            'lhs-lexical-scope': None,
            # jason data could be the boolean value true or false -> string
            'rhs-value': str(data if data is not None else json.dumps(step['value'])),
            'rhs-binary': binary_as_bytes(step['value'].get('binary'))
        }
    }

def parse_json_function_call(step, root=None):
    """Parse a function call step of a json trace."""

    return {
        'kind': 'function-call',
        'location': srcloct.json_srcloc(
            step.get('sourceLocation'), root
        ),
        'detail': {
            'name': step['function']['displayName'],
            'name-path': step['function']['identifier'],
            'location': srcloct.json_srcloc(
                step['function']['sourceLocation'], root
            )
        }
    }

def parse_json_function_return(step, root=None):
    """Parse a function return step of a json trace."""

    return {
        'kind': 'function-return',
        'location': srcloct.json_srcloc(
            step.get('sourceLocation'), root
        ),
        'detail': {
            'name': step['function']['displayName'],
            'name-path': step['function']['identifier'],
            'location': srcloct.json_srcloc(
                step['function']['sourceLocation'], root
            )
        }
    }

def parse_json_location_only(step, root=None):
    """Parse (ignore) a location-only step of a json trace."""

    _ = step
    _ = root

def binary_as_bytes(binary):
    """Reformat binary string as a sequence of bytes."""

    if not binary:
        return binary
    bits = re.sub(r'\s', '', binary)
    bites = re.findall('[01]{8}', bits)
    if bits != ''.join(bites):
        return binary
    return ' '.join(bites)

################################################################

def close_function_stack_frames(trace):
    """Append function-return steps missing from end of a trace.

    The json and xml traces from cbmc include function-call
    and function-return steps, but each error trace ends with
    a failure and omits the function returns for the function
    calls remaining on the call stack.  This appends these
    missing function returns to the end of the trace so that
    all function calls are properly nested and bracketed with
    call/return steps."""

    stack = []

    def push_stack(stack, elt):
        stack.append(elt)
        return stack

    def pop_stack(stack):
        assert stack
        return stack[-1], stack[:-1]

    location = None
    for step in trace:
        kind = step['kind']
        location = step['location']
        callee_name = step.get('detail', {}).get('name')
        callee_name_path = step.get('detail', {}).get('name-path')
        callee_location = step.get('detail', {}).get('location')

        if kind == 'function-call':
            stack = push_stack(stack, (callee_name, callee_name_path, callee_location))
            continue

        if kind == 'function-return':
            pair, stack = pop_stack(stack)
            callee_name_, _, _ = pair
            if callee_name != callee_name_:
                raise UserWarning(f'Function call-return mismatch: {callee_name} {callee_name_}')
            continue

    stack.reverse()
    for callee_name, callee_name_path, callee_location in stack:
        function_return = {
            "detail": {
                "location": callee_location,
                "name": callee_name,
                "name-path": callee_name_path
            },
            "kind": "function-return",
            "location": location,
            "hidden": True # TODO: should match corresponding call
        }
        trace.append(function_return)

    return trace

################################################################

def strip_external_srclocs(trace):
    """Strip source locations pointing to code outside the source tree.

    It is possible for a source location to point to code outside of
    the source tree (like an inlined function definition in a system
    header file).  This function replaces all such external source
    locations in the trace with the special 'missing' source location.
    """

    for step in trace:

        # The source locations produced by srcloct have paths
        # relative to the source root for all files under the
        # root, and have absolute paths for all other files.

        if step['location']['file'].startswith('/'):
            step['location'] = srcloct.MISSING_SRCLOC
        if step['detail'].get('location'):
            if step['detail']['location']['file'].startswith('/'):
                step['detail']['location'] = srcloct.MISSING_SRCLOC
    return trace

################################################################

def visible_steps(trace):
    """A generator for the visible steps appearing in a trace.

    A hidden step is generally an internal step of CBMC that is of no
    interest to a user debugging a code issue raised by CBMC, but
    there are two exceptions:

    1. Function invocations and returns should be visible.  CBMC marks
       malloc invocation as visible and return as hidden.

    2. Initialization of static data should be visible.  CBMC marks
       assignments initializing static data within CBMC initialization
       as hidden.

    This generator iterates through the visible steps after correcting
    these mischaracterizations of hidden and visible steps.
    """

    initializing = True
    for step in trace:
        if step is None:
            continue

        # make function invocation and return visible
        if step['kind'] == 'function-call':
            if step['detail']['name'] == '__CPROVER_initialize':
                initializing = True
            yield step
            continue
        if step['kind'] == 'function-return':
            if step['detail']['name'] == '__CPROVER_initialize':
                initializing = False
            yield step
            continue

        # make initialization visible
        if step['kind'] == 'variable-assignment' and initializing:
            # skip internal initialization
            if step['detail']['lhs'].startswith('__CPROVER'):
                continue
            yield step
            continue

        # Make visible steps visible
        if not step['hidden']:
            yield step

################################################################
# make-trace

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_trace(args):
    """Implementation of make-trace"""

    viewer_trace, cbmc_trace, srcdir, wkdir = (
        args.viewer_trace, args.result, args.srcdir, args.wkdir)

    if viewer_trace:
        if filet.all_json_files(viewer_trace):
            return TraceFromJson(viewer_trace)
        fail(f"Expected json files: {viewer_trace}")

    if cbmc_trace and srcdir:
        if filet.all_text_files(cbmc_trace):
            if wkdir:
                return TraceFromCbmcText(cbmc_trace, srcdir, wkdir)
            fail("Expected --srcdir, --wkdir, and cbmc trace output.")
        if filet.all_json_files(cbmc_trace):
            return TraceFromCbmcJson(cbmc_trace, srcdir)
        if filet.all_xml_files(cbmc_trace):
            return TraceFromCbmcXml(cbmc_trace, srcdir)
        fail(f"Expected json files or xml files, not both: {cbmc_trace}")

    return Trace()

def make_and_save_trace(args, path=None):
    """Make trace object and write to file or stdout"""

    obj = make_trace(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Proof summary."""

import voluptuous
import voluptuous.humanize

from cbmc_viewer import templates
from cbmc_viewer import util

################################################################
# Data passed to jinja to generate proof summary from summary.jinja.html

VALID_SUMMARY_DATA = voluptuous.Schema({
    # coverage section of the summary
    'coverage': {
        'overall': voluptuous.Any(
            {
                'percentage': float, # percentage of lines hit
                'hit': int, # lines hit
                'total': int # lines total
            },
            {}
        ),
        'function': [{
            'percentage': float, # percentage of lines hit
            'hit': int, # lines hit
            'total': int, # lines total
            'file_name': str,
            'func_name': str,
            'line_num': int,

        }]
    },
    # warning section of the summary
    'warnings': {
        'expected_missing_function': [str], # function names
        'unexpected_missing_function': [str], # function names
        'other': [str] # warning messages
    },
    # failure sections of the summary
    'failures': {
        'property': [{ # assertion failures
            'prop_name': str,
            'prop_desc': str,
            'file_name': str,
            'func_name': str,
            'line_num': int,
            'func_line': int
        }],
        'loop': [{ # loop unwinding assertion failures
            'loop_name': str,
            'file_name': str,
            'func_name': str,
            'line_num': int
        }],
        'other': [str] # unrecognized failure names
    }
}, required=True)

VALID_SUMMARY = voluptuous.Schema({
    'summary': VALID_SUMMARY_DATA,
    'outdir': str # default output directory for dump()
}, required=True)

################################################################

class Summary:
    """Proof summary.

    The summary merges results produced by cbmc into a single report.
    """

    def __init__(self, coverage, symbols, results, properties, loops,
                 config, outdir='.'):
        # pylint: disable=too-many-arguments

        self.summary = {
            'coverage': {
                'overall': overall_coverage(coverage),
                'function': function_coverage(coverage, symbols)
            },
            'warnings': {
                'expected_missing_function':
                    expected_missing_functions(results, config),
                'unexpected_missing_function':
                    unexpected_missing_functions(results, config),
                'other': other_warnings(results)
            },
            'failures': {
                'property': property_failures(results, properties, symbols),
                'loop': loop_failures(results, loops),
                'other': other_failures(results, properties, loops)
            }
        }
        self.outdir = outdir
        self.validate()

    def __str__(self):
        """Render the proof summary as html."""

        return templates.render_summary(self.summary)

    def validate(self):
        """Validate members of a summary object."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_SUMMARY
        )

    def dump(self, filename=None, outdir=None):
        """Write the proof summary to a file rendered as html."""

        util.dump(self, filename or "index.html", outdir or self.outdir)

################################################################
# Coverage data

def overall_coverage(coverage):
    """Overall proof coverage."""

    cov = coverage.overall_coverage
    if not cov:
        return {}

    return {
        'percentage': cov['percentage'],
        'hit': cov['hit'],
        'total': cov['total']
    }

def function_coverage(coverage, symbols):
    """Function proof coverage."""

    return [
        {
            'percentage': func_cov['percentage'],
            'hit': func_cov['hit'],
            'total': func_cov['total'],
            'file_name': file_name,
            'func_name': func_name,
            'line_num': symbols.lookup(func_name).get("line") if symbols.lookup(func_name) else 0
        }
        for file_name, file_data in coverage.function_coverage.items()
        for func_name, func_cov in file_data.items()
    ]

################################################################
# Warning data

# TODO: This classification of warnings should be done by make-results

def warnings(results):
    """Proof warnings."""

    prefixes = ["**** WARNING:", "warning:"]

    def strip_prefixes(string, prefixes):
        """Strip warning prefixes from a warning string."""

        for prefix in prefixes:
            if string.startswith(prefix):
                return string[len(prefix):].strip()
        return string

    return [strip_prefixes(warning, prefixes) for warning in results.warning]

def missing_functions(messages):
    """Names of missing functions."""

    prefix = "no body for function"
    length = len(prefix)
    return [warning[length:].strip() for warning in messages
            if warning.startswith(prefix)]

def expected_missing_functions(results, config):
    """Names of missing functions expected to be missing."""

    return [
        function for function in missing_functions(warnings(results))
        if function in config.expected_missing_functions()
    ]

def unexpected_missing_functions(results, config):
    """Names of missing functions not expected to be missing."""

    return [
        function for function in missing_functions(warnings(results))
        if function not in config.expected_missing_functions()
    ]

def other_warnings(results):
    """Warnings unrelated to missing functions."""

    prefix = "no body for function"
    return [warning.strip() for warning in warnings(results)
            if not warning.startswith(prefix)]

################################################################
# Failure data

def property_definition(prop_name, properties, symbols):
    """Details for a property failure."""

    prop_def = properties.lookup(prop_name)
    if prop_def is None:
        return None

    srcloc = prop_def['location']
    return {
        'prop_name': prop_name,
        'prop_desc': prop_def['description'],
        'file_name': srcloc['file'],
        'func_name': srcloc['function'] or '',
        'line_num': srcloc['line'],
        'func_line':
            # Symbol table won't contain functions modeled by CBMC (eg, strcmp)
            # Return a line number 0 for such functions
            (symbols.lookup(srcloc['function']) or {'line': 0})['line']
    }

def loop_definition(prop_name, loops):
    """Details for a loop unwinding assertion failure."""

    srcloc = loops.lookup_assertion(prop_name)
    if srcloc is None:
        return None

    return {
        'loop_name': prop_name,
        'file_name': srcloc['file'],
        'func_name': srcloc['function'],
        'line_num': srcloc['line']
    }

def other_definition(prop_name, properties, loops):
    """Name of an unrecognized failure."""

    if properties.lookup(prop_name) or loops.lookup_assertion(prop_name):
        return None
    return prop_name

def filter_none(items):
    """Remove instances of None from a list of items."""

    return [item for item in items if item is not None]

def property_failures(results, properties, symbols):
    """Details for property failures."""

    failures = results.results[False]
    return filter_none(
        [property_definition(failure, properties, symbols)
         for failure in failures]
    )

def loop_failures(results, loops):
    """Details for loop unwinding assertion failures."""

    failures = results.results[False]
    return filter_none(
        [loop_definition(failure, loops) for failure in failures]
    )

def other_failures(results, properties, loops):
    """Names of unrecognized failures."""

    failures = results.results[False]
    return filter_none(
        [other_definition(failure, properties, loops) for failure in failures]
    )

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Report the results of cbmc.

Report the results of cbmc with annotated source files indiciating
line coverage, with coverage reports for statically reachable functions,
and with lists of property violations and traces for each violation.
"""

from pathlib import Path
import datetime
import logging
import os
import sys

from cbmc_viewer import configt
from cbmc_viewer import coveraget
from cbmc_viewer import loopt
from cbmc_viewer import propertyt
from cbmc_viewer import reachablet
from cbmc_viewer import report
from cbmc_viewer import resultt
from cbmc_viewer import sourcet
from cbmc_viewer import symbolt
from cbmc_viewer import tracet

################################################################

class Timer:
    """A simple timer for display in progress information."""

    def __init__(self):
        self.time = datetime.datetime.now()

    def reset(self):
        """Reset timer to current time."""

        self.time = datetime.datetime.now()

    def elapsed(self):
        """Reset timer and return time elapsed since last reset."""

        old = self.time
        self.time = datetime.datetime.now()
        return (self.time - old).total_seconds()

TIMER = Timer()

def progress(msg, done=False):
    """Display local progress for a single step."""

    if done:
        time = TIMER.elapsed()
        logging.info("viewer: %s...done (%s sec)", msg, time)
    else:
        TIMER.reset()
        logging.info("viewer: %s...", msg)

GLOBAL_TIMER = Timer()

def global_progress(msg, done=False):
    """Display overall progress for all steps."""

    if done:
        time = GLOBAL_TIMER.elapsed()
        logging.info("viewer: %s...done (%s sec)", msg, time)
    else:
        GLOBAL_TIMER.reset()
        logging.info("viewer: %s...", msg)

################################################################

def viewer(args):
    """Construct the cbmc report."""

    if not (args.result or args.viewer_result or args.coverage or args.viewer_coverage):
        logging.error("Need property checking results or coverage checking results.")
        sys.exit(1)

    global_progress("CBMC viewer")

    htmldir = os.path.join(args.reportdir, "html")
    jsondir = os.path.join(args.reportdir, "json")
    os.makedirs(htmldir, exist_ok=True)
    os.makedirs(jsondir, exist_ok=True)
    jsondir = Path(jsondir)

    progress("Scanning property checking results")
    results = resultt.make_and_save_result(args, jsondir / 'viewer-result.json')
    progress("Scanning property checking results", True)

    progress("Scanning error traces")
    traces = tracet.make_and_save_trace(args, jsondir / 'viewer-trace.json')
    progress("Scanning error traces", True)

    progress("Scanning coverage data")
    coverage = coveraget.make_and_save_coverage(args, jsondir / 'viewer-coverage.json')
    progress("Scanning coverage data", True)

    progress("Scanning loop definitions")
    loops = loopt.make_and_save_loop(args, jsondir / 'viewer-loop.json')
    progress("Scanning loop definitions", True)

    progress("Scanning properties")
    properties = propertyt.make_and_save_property(args, jsondir / 'viewer-property.json')
    progress("Scanning properties", True)

    progress("Computing reachable functions")
    reachablet.make_and_save_reachable(args, jsondir / 'viewer-reachable.json')
    progress("Computing reachable functions", True)

    # Make sources last, it may delete the goto binary
    progress("Scanning source tree")
    sources = sourcet.make_and_save_source(args, jsondir / 'viewer-source.json')
    progress("Scanning source tree", True)

    progress("Preparing symbol table")
    symbols = symbolt.make_and_save_symbol(args, jsondir / 'viewer-symbol.json')
    progress("Preparing symbol table", True)

    config = configt.Config(args.config)
    report.report(config, sources, symbols, results, coverage, traces,
                  properties, loops, htmldir, progress)

    global_progress("CBMC viewer", True)
    return 0 # exit with normal return code



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC properties checked during property checking.

The Result module describes the results of CBMC property checking, the
Coverage module describes the results of CBMC coverage checking, and
the Property module describes the properties checked during property
checking.
"""

import json
import logging

import voluptuous
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import srcloct
from cbmc_viewer import util

JSON_TAG = "viewer-property"

################################################################
# Property validator

VALID_PROPERTY_DEFINITION = voluptuous.Schema({
    'class': str,       # eg, "pointer dereference"
    'description': str, # eg, "pointer outside dynamic object"
    'expression': str,
    'location': srcloct.VALID_SRCLOC
}, required=True)

VALID_PROPERTY = voluptuous.Schema({
    'properties': {
        # property name -> property definition
        voluptuous.Optional(str): VALID_PROPERTY_DEFINITION
    }
}, required=True)

################################################################

def key(name):
    """A key for sorting property names like function.property.index"""

    try:
        dot = name.rindex('.')
        return (name[:dot].lower(), int(name[dot+1:]))
    except ValueError:
        logging.warning("Property name not of the form STRING.INTEGER: %s",
                        name)
        return (name, 0)

################################################################

class Property:
    """CBMC properties checked during property checking."""

    def __init__(self, property_lists=None):
        """Load CBMC properties from lists of properties."""

        property_lists = property_lists or []

        def handle_duplicates(name, defn1, defn2):
            logging.warning("Found duplicate property definition: "
                            "%s: %s %s <- %s %s",
                            name,
                            defn1["location"]["file"],
                            defn1["location"]["line"],
                            defn2["location"]["file"],
                            defn2["location"]["line"])

        self.properties = util.merge_dicts(property_lists, handle_duplicates)
        self.validate()

    def __repr__(self):
        """A dict representation of an property table."""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of an property table."""

        return json.dumps({JSON_TAG: self.__repr__()}, indent=2, sort_keys=True)

    def validate(self):
        """Validate properties."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_PROPERTY
        )

    def dump(self, filename=None, directory=None):
        """Write properties to a file or stdout."""

        util.dump(self, filename, directory)

    def names(self):
        """Names of known properties."""

        return self.properties.keys()

    def lookup(self, name):
        """Look up all known information for named property."""

        return self.properties.get(name)

    def get_description(self, name):
        """Get description of named property."""

        return (self.lookup(name) or {}).get('description') or name

    def get_srcloc(self, name):
        """Get source location of named property."""

        return (self.lookup(name) or {}).get('location')

################################################################

class PropertyFromJson(Property):
    """Load property table from the output of make-property.

    Given a list of json files containing property definitions
    produced by make-property, merge these lists into a single list
    of properties.
    """

    def __init__(self, json_files):

        super().__init__(
            [parse.parse_json_file(json_file)[JSON_TAG]["properties"]
             for json_file in json_files]
        )

################################################################

class PropertyFromCbmcJson(Property):
    """Load properties from output of 'cbmc --show-properties --json-ui'."""

    def __init__(self, json_files, root):

        super().__init__(
            [load_cbmc_json(json_file, root) for json_file in json_files]
        )

def load_cbmc_json(jsonfile, root):
    """Load a json file produced by cbmc --show-properties --json-ui."""

    json_data = parse.parse_json_file(jsonfile)
    assert json_data is not None

    # Search cbmc output for {"properties": [ PROPERTY ]}
    asserts = [json_map for json_map in json_data if "properties" in json_map]
    if len(asserts) != 1:
        raise UserWarning(f"Expected 1 set of properties in cbmc output, found {len(asserts)}")

    # Each PROPERTY a loop property and definition
    root = srcloct.abspath(root)
    return {
        property['name']: {
            'class': property['class'],
            'description': property['description'],
            'expression': property['expression'],
            'location': srcloct.json_srcloc(property['sourceLocation'], root)
        }
        for property in asserts[0]["properties"]
    }

################################################################

class PropertyFromCbmcXml(Property):
    """Load properties from output of 'cbmc --show-properties --xml-ui'."""

    def __init__(self, xml_files, root):

        super().__init__(
            [load_cbmc_xml(xml_file, root) for xml_file in xml_files]
        )

def load_cbmc_xml(xmlfile, root):
    """Load a json file produced by cbmc --show-properties --xml-ui."""

    xml_data = parse.parse_xml_file(xmlfile)
    assert xml_data is not None

    root = srcloct.abspath(root)
    return {
        property.get("name"): {
            "class": property.get("class"),
            "description": property.find("description").text,
            "expression": property.find("expression").text,
            'location': srcloct.xml_srcloc(property.find("location"), root)
        }
        for property in xml_data.iter("property")
    }

################################################################
# make-property

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_property(args):
    """The implementation of make-property."""

    viewer_property, cbmc_property, srcdir = args.viewer_property, args.property, args.srcdir

    if viewer_property:
        if filet.all_json_files(viewer_property):
            return PropertyFromJson(viewer_property)
        fail(f"Expected json files: {viewer_property}")

    if cbmc_property and srcdir:
        if filet.all_json_files(cbmc_property):
            return PropertyFromCbmcJson(cbmc_property, srcdir)
        if filet.all_xml_files(cbmc_property):
            return PropertyFromCbmcXml(cbmc_property, srcdir)
        fail(f"Expected json files or xml files, not both: {cbmc_property}")

    logging.info("make-property: nothing to do: need "
                 "cbmc property listing results (cbmc --show-properties) and --srcdir or "
                 "--viewer-property")
    return Property()

def make_and_save_property(args, path=None):
    """Make property object and write to file or stdout"""

    obj = make_property(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Assemble the full report for cbmc viewer."""

import logging
import os
import shutil

import pkg_resources

from cbmc_viewer import markup_code
from cbmc_viewer import markup_summary
from cbmc_viewer import markup_trace

PACKAGE = 'cbmc_viewer'
VIEWER_JS = 'viewer.js'
VIEWER_CSS = 'viewer.css'

def progress_default(string):
    """A default method for logging progress."""

    logging.info(string)

def report(config, sources, symbols, results, coverage, traces, properties,
           loops, report_dir='.', progress=progress_default):
    """Assemble the full report for cbmc viewer."""

    # The report is assembled from many sources of data
    # pylint: disable=too-many-locals

    # Some code depends on these definitions
    #   * links to traces in summary produced with jinja summary template
    #   * links to sources in traces produced by markup_trace
    #   * links to css and js in traces produced with jinja trace template
    code_dir = report_dir
    trace_dir = os.path.join(report_dir, markup_trace.TRACES)

    os.makedirs(report_dir, exist_ok=True)
    shutil.copy(pkg_resources.resource_filename(PACKAGE, VIEWER_CSS),
                report_dir)
    shutil.copy(pkg_resources.resource_filename(PACKAGE, VIEWER_JS),
                report_dir)

    progress("Preparing report summary")
    markup_summary.Summary(
        coverage, symbols, results, properties, loops, config).dump(
            outdir=report_dir)
    progress("Preparing report summary", True)

    progress("Annotating source tree")
    for path in sources.files:
        markup_code.Code(sources.root, path, symbols, coverage).dump(
            outdir=code_dir)
    progress("Annotating source tree", True)

    progress("Annotating traces")
    snippets = markup_trace.CodeSnippet(sources.root)
    for name, trace in traces.traces.items():
        markup_trace.Trace(name, trace, symbols, properties, loops, snippets).dump(
            outdir=trace_dir)
    progress("Annotating traces", True)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC reachable functions"""

import json
import logging
import os
import subprocess
import tempfile

import voluptuous
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import runt
from cbmc_viewer import srcloct
from cbmc_viewer import util

JSON_TAG = 'viewer-reachable'

################################################################
# Reachable validator

VALID_REACHABLE = voluptuous.Schema({
    'reachable': {
        # file name -> list of reachable functions
        voluptuous.Optional(str): [str]
    }
}, required=True)

################################################################

class Reachable:
    """CBMC reachable functions."""

# TODO: again dicts -> dict

    def __init__(self, function_lists=None):
        """Load CBMC reachable functions from lists of functions."""

        function_lists = function_lists or []

        functions = self.merge_function_lists(function_lists)
        self.reachable = self.sort_function_names(functions)
        self.validate()

    def __repr__(self):
        """A dict representation of the reachable functions."""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of the reachable functions."""

        return json.dumps({JSON_TAG: self.__repr__()}, indent=2, sort_keys=True)

    def validate(self):
        """Validate reachable functions."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_REACHABLE
        )

    def dump(self, filename=None, directory=None):
        """Write reachable functions to a file or stdout."""

        util.dump(self, filename, directory)

    @staticmethod
    def merge_function_lists(function_lists):
        """Merge lists of functions into a single list."""

        if len(function_lists) == 1:
            return function_lists[0]

        functions = {}
        for function_list in function_lists:
            for file_name, func_names in function_list.items():
                functions[file_name] = functions.get(file_name, set())
                functions[file_name] = functions[file_name].union(func_names)

        return functions

    @staticmethod
    def sort_function_names(functions):
        """Sort the names of reachable functions in a file."""

        return {file_name: sorted(func_names)
                for file_name, func_names in functions.items()}

################################################################

class ReachableFromJson(Reachable):
    """Load the reachable functions from the output of make-reachable.

    Given a list of json files containing symbol tables produced by
    make-reachable, merge these function lists into a single list.
    """

    def __init__(self, json_files):

        super().__init__(
            [parse.parse_json_file(json_file)[JSON_TAG]["reachable"]
             for json_file in json_files]
        )

################################################################

class ReachableFromCbmcJson(Reachable):
    """Load functions from output of goto-analyzer --reachable-functions --json.
    """

    def __init__(self, json_files, root):

        super().__init__(
            [load_cbmc_json(json_file, root) for json_file in json_files]
        )

def load_cbmc_json(json_file, root):
    """Load json file produced by goto-analyzer --reachable-functions --json json_file"""

    json_data = parse.parse_json_file(json_file)
    return parse_cbmc_json(json_data, root)

# TODO: Does file-local name mangling mess this up?

def parse_cbmc_json(json_data, root):
    """Parse json output of goto-analyzer --reachable-functions --json."""

    root = srcloct.abspath(root)
    reachable = {}
    for function in json_data:
        func_name = function['function']
        if func_name.startswith('__CPROVER'):
            continue

        # goto-analyzer --reachable-functions produces a malformed
        # entry for __CPROVER_Start with an empty string for the
        # function name.
        if not func_name:
            continue

        file_name = function.get('file')
        if file_name is None:
            logging.warning('Skipping reachable function with invalid source location: %s',
                            function)
            continue

        file_name = srcloct.abspath(file_name)
        if srcloct.is_builtin(file_name):
            continue
        if not file_name.startswith(root):
            continue

        path = srcloct.relpath(file_name, root)
        reachable[path] = reachable.get(path, set())
        reachable[path].add(func_name)

    return reachable

################################################################

class ReachableFromCbmcXml(Reachable):
    """Load functions from output of goto-analyzer --reachable-functions --xml.
    """

    def __init__(self, xml_files, root):

        super().__init__([])

        _, _ = xml_files, root
        raise UserWarning("The goto-analyzer --xml option generates text "
                          "and not xml.")

################################################################

class ReachableFromGoto(Reachable):
    """Load reachable functions of a goto binary."""

    def __init__(self, goto, root, cwd=None):

        # pylint: disable=consider-using-with
        data = tempfile.NamedTemporaryFile(delete=False)

        cmd = ["goto-analyzer", "--reachable-functions", "--json", data.name, goto]
        try:
            runt.run(cmd, cwd=cwd)
        except subprocess.CalledProcessError as err:
            raise UserWarning(f'Failed to run {cmd}: {err}') from err

        json_data = parse.parse_json_file(data.name)
        os.unlink(data.name)

        super().__init__(
            [parse_cbmc_json(json_data, root)]
        )

################################################################
# make-reachable

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_reachable(args):
    """The implementation of make-reachable."""

    viewer_reachable, cbmc_reachable, srcdir, goto = (
        args.viewer_reachable, None, args.srcdir, args.goto)

    if viewer_reachable:
        if filet.all_json_files(viewer_reachable):
            return ReachableFromJson(viewer_reachable)
        fail(f"Expected json files: {viewer_reachable}")

    if cbmc_reachable and srcdir:
        if filet.all_json_files(cbmc_reachable):
            return ReachableFromCbmcJson(cbmc_reachable, srcdir)
        if filet.all_xml_files(cbmc_reachable):
            return ReachableFromCbmcXml(cbmc_reachable, srcdir)
        fail(f"Expected json files or xml files, not both: {cbmc_reachable}")

    if goto and srcdir:
        return ReachableFromGoto(goto, srcdir)

    logging.info("make-reachable: nothing to do: need "
                 "--srcdir and --goto, or "
                 "cbmc reachable functions results (goto-analyzer --reachable-functions)"
                 " and --srcdir, or "
                 "--viewer-reachable")
    return Reachable()

def make_and_save_reachable(args, path=None):
    """Make reachable object and write to file or stdout"""

    obj = make_reachable(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC loop information"""

import json
import logging
import re
import subprocess

import voluptuous
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import runt
from cbmc_viewer import srcloct
from cbmc_viewer import util

JSON_TAG = 'viewer-loop'

################################################################
# Loop validator

VALID_LOOP = voluptuous.Schema({
    'loops': {
        # loop name -> loop srcloc
        voluptuous.Optional(str): srcloct.VALID_SRCLOC
    }
}, required=True)

################################################################

class Loop:
    """CBMC loop information."""

    def __init__(self, loop_maps=None):
        """Load CBMC loop map from a lists of loop maps."""

        loop_maps = loop_maps or []

        def handle_duplicates(name, srcloc1, srcloc2):
            logging.warning("Found duplicate loop definition: "
                            "%s: %s %s <- %s %s",
                            name,
                            srcloc1["file"],
                            srcloc1["line"],
                            srcloc2["file"],
                            srcloc2["line"])

        self.loops = util.merge_dicts(loop_maps, handle_duplicates)
        self.loops = self.unmangle_loop_names(self.loops)
        self.validate()

    def __repr__(self):
        """A dict representation of the loop table."""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of the loop table."""

        return json.dumps({JSON_TAG: self.__repr__()}, indent=2, sort_keys=True)

    def validate(self):
        """Validate loops."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_LOOP
        )

    def dump(self, filename=None, directory=None):
        """Write loops to a file or stdout."""

        util.dump(self, filename, directory)


# TODO: generally don't fail silently, give warning or raise exception
# TODO: generally give hints for how to fix the problem (give actions to take)

    @staticmethod
    def unmangle_loop_names(loops):
        """Reverse CBMC name mangling of loop names in static functions."""

        # Loops in static function FUNC may have names FUNC$link1.0
        static_names = [name for name in loops if '$link' in name]

        # Replace the name FUNC$link1.0 with FUNC.0
        for long_name in static_names:
            long_func, index = long_name.split('.')
            short_func = long_func.split('$')[0]
            short_name = f'{short_func}.{index}'
            assert long_name != short_name

            if short_name in loops:
                if loops[short_name] != loops[long_name]:
                    logging.warning("Found duplicate loop definition: "
                                    "%s: %s %s <- %s %s",
                                    short_name,
                                    loops[short_name]["file"],
                                    loops[short_name]["line"],
                                    loops[long_name]["file"],
                                    loops[long_name]["line"])
                continue
            loops[short_name] = loops[long_name]
            del loops[long_name]
        return loops

    def names(self):
        """Loop names."""

        return self.loops.keys()

    def lookup(self, name):
        """Look up the srcloc for the named loop."""

        return self.loops.get(name)

    def lookup_static(self, name):
        """Look up the srcloc for the named loop in a static function."""

        # CBMC property checking refers to loop K in function FCN as FCN.K,
        # but the goto-cc --export-file-local-symbols flag renames FCN.K
        # appearing in a file FILE in a static function FCN as
        # __CPROVER_file_local_FILE_FCN_LOOP.N
        keys = [key for key in self.loops.keys() if key.endswith('_'+name)]
        if not keys:
            return None
        if len(keys) != 1:
            raise UserWarning(f"Loop name {name} matches {len(keys)} static loop names: {keys}")

        return self.loops.get(keys[0])

    def lookup_assertion(self, name):
        """Look up the srcloc for the named loop unwinding assertion."""

        # CBMC refers to the loop K in function FCN as FCN.K and to
        # the unwinding assertion associated with that loop as
        # FCN.unwind.K
        match = re.match(r'^(.*)\.unwind\.([0-9]+)$', name)
        if match is None:
            return None
        loop = f'{match.group(1)}.{match.group(2)}'

        return self.lookup(loop) or self.lookup_static(loop)

################################################################

class LoopFromJson(Loop):
    """Load cbmc loop table from output of make-loop.

    Given a list of json files containing symbol tables produced by
    make-loop, merge these loop tables into a single loop table.
    """


    def __init__(self, json_files):

        super().__init__(
            [parse.parse_json_file(json_file)[JSON_TAG]['loops']
             for json_file in json_files]
        )

################################################################

class LoopFromCbmcJson(Loop):
    """Load loop table from output of 'cbmc --show-loops --json-ui."""

    def __init__(self, json_files, root):

        super().__init__(
            [load_cbmc_json(json_file, root) for json_file in json_files]
        )

def load_cbmc_json(json_file, root):
    """Load a json file produced by cbmc --show-loops --json-ui."""

    return parse_cbmc_json(parse.parse_json_file(json_file), root)

def parse_cbmc_json(json_data, root):
    """Parse the json output of cbmc --show-loops --json-ui."""

    # Search cbmc output for {"loops": [ LOOP ]}
    loops = [json_map for json_map in json_data if "loops" in json_map]
    if len(loops) != 1:
        raise UserWarning(f"Expected 1 set of loops in cbmc output, found {len(loops)}")

    # Each LOOP is a dict that gives a loop name and location.
    root = srcloct.abspath(root)
    return {loop['name']:
            srcloct.json_srcloc(loop['sourceLocation'], root)
            for loop in loops[0]["loops"]}

################################################################

class LoopFromCbmcXml(Loop):
    """Load loop table from output of 'cbmc --show-loops --xml-ui."""

    def __init__(self, xml_files, root):

        super().__init__(
            [load_cbmc_xml(xml_file, root) for xml_file in xml_files]
        )

def load_cbmc_xml(xml_file, root):
    """Load an xml file produced by cbmc --show-loops --xml-ui."""

    return parse_cbmc_xml(parse.parse_xml_file(xml_file), root)

def parse_cbmc_xml(xml_data, root):
    """Parse the xml output of cbmc --show-loops --xml-ui."""

    # Each loop element contains loop-id and location elements
    root = srcloct.abspath(root)
    return {loop.find("loop-id").text:
            srcloct.xml_srcloc(loop.find("location"), root)
            for loop in xml_data.iter("loop")}

################################################################

class LoopFromGoto(Loop):
    """Load loop table from a goto binary."""

    def __init__(self, goto, root, cwd=None):

        cmd = ['cbmc', '--show-loops', '--json-ui', goto]
        try:
            super().__init__(
                [parse_cbmc_json(json.loads(runt.run(cmd, cwd=cwd)), root)]
            )
        except subprocess.CalledProcessError as err:
            raise UserWarning(f'Failed to run {cmd}: {err}') from err
        except json.decoder.JSONDecodeError as err:
            raise UserWarning(f'Failed to parse output of {cmd}: {err}') from err

################################################################
# make-loop

# pylint: disable=inconsistent-return-statements

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_loop(args):
    """The implementation of make-loop"""

    viewer_loop, cbmc_loop, srcdir, goto = args.viewer_loop, None, args.srcdir, args.goto

    if viewer_loop:
        if filet.all_json_files(viewer_loop):
            return LoopFromJson(viewer_loop)
        fail(f"Expected json files: {viewer_loop}")

    if cbmc_loop and srcdir:
        if filet.all_json_files(cbmc_loop):
            return LoopFromCbmcJson(cbmc_loop, srcdir)
        if filet.all_xml_files(cbmc_loop):
            return LoopFromCbmcXml(cbmc_loop, srcdir)
        fail(f"Expected json files or xml files, not both: {cbmc_loop}")

    if goto and srcdir:
        return LoopFromGoto(goto, srcdir)

    logging.info("make-loop: nothing to do: need "
                 "--goto and --srcdir, or "
                 "cbmc loop listing results (cbmc --show-loops) and --srcdir, or "
                 "--viewer-loop")
    return Loop()

def make_and_save_loop(args, path=None):
    """Make loop object and write to file or stdout"""

    obj = make_loop(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Links to source code.

This module is a set of methods for constructing links into the
annotated source code.  All other modules use these methods for
consistent links to source code.  All paths in this module are
assumed to be relative to the root of the source code.

"""

import html
import os
import re

from cbmc_viewer import srcloct

################################################################

def path_to_file(dst, src):
    """The path from src to dst for use in a hyperlink from src to dst.

    Given two paths src and dst relative to a common root, return the
    relative path from src to dst.  This is the path to use in a
    hyperlink from src to dst.  For example,
    the path from 'a/b/foo.html' to 'c/bar.html' is '../../c/bar.html',
    the path from 'a/b/foo.html' to '.' is '../..', and
    the path from '.' to 'a/b/foo.html' is 'a/b/foo.html'.
    """

    src_dir = os.path.dirname(src)
    src_to_dst = os.path.relpath(dst, src_dir)
    cwd_to_dst = os.path.normpath(os.path.join(src_dir, src_to_dst))
    if dst != cwd_to_dst:
        raise UserWarning(f"{dst} != {cwd_to_dst}")
    return src_to_dst

################################################################
# Method to link into the source tree.
# By default, links are from the root of the source tree to the source file.
#
# In what follows, the text being linked into the source tree may be
# None.  For example, the source location for a global variable (a
# static variable) will give a file name and a line number, but the
# function name will be omitted (because there is no enclosing
# function to name).  This may result in one of the following
# functions being invoked with text set to None as the value of the
# missing function name.

def link_text_to_file(text, to_file, from_file=None, escape_text=True):
    """Link text to a file in the source tree."""

    if not text:
        return text

    text = html.escape(str(text)) if escape_text else str(text)

    if srcloct.file_is_not_a_source_file(to_file):
        return text

    from_file = from_file or '.'
    path = path_to_file(to_file, from_file)
    return f'<a href="{path}.html">{text}</a>'

def link_text_to_line(text, to_file, line, from_file=None, escape_text=True):
    """Link text to a line in a file in the source tree."""

    if not text:
        return text

    text = html.escape(str(text)) if escape_text else str(text)

    if srcloct.file_is_not_a_source_file(to_file):
        return text

    from_file = from_file or '.'
    line = int(line)
    path = path_to_file(to_file, from_file)
    return f'<a href="{path}.html#{line}">{text}</a>'

def link_text_to_srcloc(text, srcloc, from_file=None, escape_text=True):
    """Link text to a source location in a file in the source tree."""

    if not text:
        return text

    if srcloc is None:
        return html.escape(text) if escape_text else text
    return link_text_to_line(text, srcloc['file'], srcloc['line'], from_file, escape_text)

def link_text_to_symbol(text, symbol, symbols, from_file=None, escape_text=True):
    """Link text to a symbol definition in the source tree."""

    if not text:
        return text

    srcloc = symbols.lookup(symbol)
    return link_text_to_srcloc(text, srcloc, from_file, escape_text=escape_text)

def split_text_into_symbols(text):
    """Split text into substrings that could be symbols."""

    return re.split('([_a-zA-Z][_a-zA-Z0-9]*)', text)

def link_symbols_in_text(text, symbols, from_file=None, escape_text=True):
    """Link symbols appearing in text to their definitions."""

    if text is None:
        return None

    tokens = split_text_into_symbols(text)
    return ''.join(
        [link_text_to_symbol(tkn, tkn, symbols, from_file, escape_text)
         for tkn in tokens]
    )

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Run a command with simple error handling."""

import logging
import subprocess
import sys

def run(cmd, cwd=None, ignored=None, encoding=None):
    """Run command cmd in directory cwd.

    The argument 'ignored' may be a list of integers giving command
    return codes that are acceptable and can be ignored.

    The argument 'encoding' is a character encoding for the text
    string captured as stdout and stderr.  The default encoding for
    modern Python is utf-8, but source code written on Windows
    platforms uses a different character encoding.  In this case,
    'latin1' is a reasonable choice, since it agrees with utf-8 on the
    ascii character set.  The wikipedia page on latin1 goes so far as
    to say latin1 is "often assumed to be the encoding of 8-bit text
    on Unix and Microsoft Windows...".
    """

    kwds = {
        'cwd': cwd,
        'stdout': subprocess.PIPE,
        'stderr': subprocess.PIPE,
        'universal_newlines': True,
    }

    # encoding keyword argument was introduced in Python 3.6
    if sys.version_info >= (3, 6):
        kwds['encoding'] = encoding

    logging.debug('run: cmd: %s', ' '.join(cmd))
    logging.debug('run: kwds: %s', kwds)

    result = subprocess.run(cmd, **kwds, check=False)

    if result.returncode:
        logging.debug('Failed to run command: %s', ' '.join(cmd))
        logging.debug('Failed return code: %s', result.returncode)
        logging.debug('Failed stdout: %s', result.stdout)
        logging.debug('Failed stderr: %s', result.stderr)
        if ignored is None or result.returncode not in ignored:
            result.check_returncode()
        logging.debug('Ignoring failure to run command: %s', cmd)

    return result.stdout

def popen(cmd, cwd=None, stdin=None, encoding=None):
    """Run a command with string stdin on stdin, return stdout and stderr."""

    cmd = [str(word) for word in cmd]
    kwds = {'cwd': cwd,
            'text': True,
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE}
    if sys.version_info >= (3, 6): # encoding is new in Python 3.6
        kwds['encoding'] = encoding or 'utf-8'
    try:
        logging.debug('Command for popen: "%s"', ' '.join(cmd))
        logging.debug('Command stdin: "%s"', stdin)
        with subprocess.Popen(cmd, **kwds) as pipe:
            stdout, stderr = pipe.communicate(input=stdin)
        logging.debug('Command stdout: "%s"', stdout)
        logging.debug('Command stderr: "%s"', stderr)
        if pipe.returncode:
            logging.debug('Command failed "%s"', ' '.join(cmd))
            logging.debug('Command return code: "%s"', pipe.returncode)
            raise UserWarning(f"Failed to run command: {' '.join(cmd)}")
        return stdout, stderr
    except FileNotFoundError as error:
        logging.debug("FileNotFoundError: command '%s': %s", ' '.join(cmd), error)
        raise UserWarning(f"Failed to run command: {' '.join(cmd)}") from error



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Command line options."""

import argparse
import logging
import os
import platform


from cbmc_viewer import filet
from cbmc_viewer import version as viewer_version
from cbmc_viewer.sourcet import Sources

from cbmc_viewer import coveraget
from cbmc_viewer import loopt
from cbmc_viewer import propertyt
from cbmc_viewer import reachablet
from cbmc_viewer import resultt
from cbmc_viewer import sourcet
from cbmc_viewer import symbolt
from cbmc_viewer import tracet
from cbmc_viewer import viewer

choices = []
if platform.system() != 'Windows':
    choices.append('find')
choices.extend(['walk', 'make', 'goto'])

OPTION_GROUPS = [
    {'group_name': "CBMC output",
     'group_desc': """
         This is the output of CBMC that is summarized by cbmc-viewer.
         Output can be text, xml, or json, but xml is strongly preferred.""",
     'group_opts': [
         {'flag': '--result',
          'metavar': 'FILE',
          'nargs': '+',
          'help': 'CBMC property checking the results.  The output of "cbmc".'},
         {'flag': '--coverage',
          'metavar': 'FILE',
          'nargs': '+',
          'help': 'CBMC coverage checking results.  The output of "cbmc --cover locations".'},
         {'flag': '--property',
          'metavar': 'FILE',
          'nargs': '+',
          'help': 'CBMC properties checked during property checking.  '
                  'The output of "cbmc --show-properties".'}]},

    {'group_name': 'Source files',
     'group_desc': None,
     'group_opts': [
         {'flag': '--srcdir',
          'type': os.path.abspath,
          'help': 'The root of the source tree, typically the root of the code repository.'},
         {'flag': '--exclude',
          'help': """
              A regular expression for the paths relative to SRCDIR to exclude from
              the list of source files.  This is rarely used."""},
         {'flag': '--extensions',
          'metavar': 'REGEXP',
          'default': r'^\.(c|h|inl)$',
          'help': """
              A regular expression for the file extensions of files to include
              in the list of source files.  This is rarely used.  (Default: %(default)s)"""},
         {'flag': '--source-method',
          'metavar': 'MHD',
          'choices': choices,
          'help': """
               The method to use to enumerate the list of source files. This is
               rarely used.  The default method 'goto' is generally to use the files
               mentioned in the symbol table of the goto binary.  The full set of
               methods available is
               1) 'find': use the Linux 'find' command in SRCDIR,
               2) 'walk': use the Python 'walk' method in SRCDIR,
               3) 'make': use the 'make' command in the WKDIR to build the goto
                   binary with the preprocessor and use the files under SRCDIR
                   mentioned in the preprocessor output, and
               4) 'goto': use the files under SRCDIR mentioned in the symbol table
                   of the goto binary."""}]},
    {'group_name': 'GOTO binaries',
     'group_desc': None,
     'group_opts': [
         {'flag': '--wkdir',
          'default': ".",
          'type': os.path.abspath,
          'help': """
              The working directory.  This is generally the directory in which
              `goto-cc` was invoked to build the goto binary.  It is the working
              directory that is mentioned in the source locations in the goto
              binary."""},
         {'flag': '--goto',
          'help': 'The goto binary itself.'}]},

    {'group_name': 'Viewer output',
     'group_desc': None,
     'group_opts': [
         {'flag': '--reportdir',
          'default': 'report',
          'type': os.path.abspath,
          'help': """
              The report directory.  Write the final report to this
              directory.  (Default: %(default)s)"""},
         {'flag': '--json-summary',
          'metavar': 'JSON',
          'help': 'Write summary of key metrics to this json file.'}]},

    {'group_name': 'Viewer input',
     'group_desc': """
         Load json output of cbmc-viewer like "viewer-coverage.json" or the
         output of cbmc-viewer subcommands like "cbmc-viewer coverage".""",
     'group_opts': [
         {'flag': '--viewer-coverage',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer coverage" giving
              CBMC coverage checking results."""},
         {'flag': '--viewer-loop',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer loop" listing the loops
              found in the goto binary."""},
         {'flag': '--viewer-property',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer property" listing the properties
              checked during CBMC property checking."""},
         {'flag': '--viewer-reachable',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer property" listing the
              reachable functions in the goto binary."""},
         {'flag': '--viewer-result',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer result" giving the
              CBMC property checking results."""},
         {'flag': '--viewer-source',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer source" listing the source
              files used to build the goto binary."""},
         {'flag': '--viewer-symbol',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer symbol" listing the symbols
              in the goto binary."""},
         {'flag': '--viewer-trace',
          'metavar': 'JSON',
          'nargs': '+',
          'help': """
              Load the output of "cbmc-viewer" or "cbmc-viewer trace" giving the error traces
              generated by CBMC for the issues found during property checking."""}]},

    {'group_name': 'Other',
     'group_desc': None,
     'group_opts': [
         {'flag': '--verbose',
          'action': 'store_true',
          'help': 'Verbose output.'},
         {'flag': '--debug',
          'action': 'store_true',
          'help': 'Debugging output.'},
         {'flag': '--version',
          'action': 'version',
          'version': viewer_version.version(),
          'help': 'Display version number and exit.'},
         {'flag': '--config',
          'metavar': 'JSON',
          'default': "cbmc-viewer.json",
          'help': """
              Viewer configuration file used (for example) to list the functions
              intentionally omitted from the verification.  (Default: '%(default)s')"""}]},

    {'group_name': 'Deprecated',
     'group_desc': 'Options from prior versions now deprecated.',
     'group_opts': [
         {'flag': '--block',
          'metavar': 'BLOCK',
          'help': 'Use --coverage instead.'},
         {'flag': '--htmldir',
          'metavar': 'HTMLDIR',
          'help': 'Use --reportdir instead.'},
         {'flag': '--srcexclude',
          'metavar': 'EXCLUDE',
          'help': 'Use --exclude instead.'},
         {'flag': '--blddir',
          'help': 'Ignored.'},
         {'flag': '--storm',
          'help': 'Ignored.'}]}
]

SUBPARSERS = [
    {'name': None,
     'func': viewer.viewer,
     'desc': 'Generate a browsable summary of CBMC results',
     'flags': None},
    {'name': 'coverage',
     'func': coveraget.make_and_save_coverage,
     'desc': 'Summarize coverage checking results',
     'flags': ['--coverage', '--viewer-coverage', '--srcdir']},
    {'name': 'loop',
     'func': loopt.make_and_save_loop,
     'desc': 'List loops found in the goto binary',
     'flags': ['--viewer-loop', '--goto', '--srcdir']},
    {'name': 'property',
     'func': propertyt.make_and_save_property,
     'desc': 'List properties checked for during property checking',
     'flags': ['--property', '--viewer-property', '--srcdir']},
    {'name': 'reachable',
     'func': reachablet.make_and_save_reachable,
     'desc': 'List reachable functions in the goto binary',
     'flags': ['--viewer-reachable', '--goto', '--srcdir']},
    {'name': 'result',
     'func': resultt.make_and_save_result,
     'desc': 'Summarize CBMC property checking results',
     'flags': ['--result', '--viewer-result']},
    {'name': 'source', # TODO: add --files back
     'func': sourcet.make_and_save_source,
     'desc': 'List source files used to build the goto binary',
     'flags': ['--viewer-source', '--goto', '--srcdir', '--wkdir',
               '--source-method', '--extensions', '--exclude']},
    {'name': 'symbol',  # TODO: add --files back
     'func': symbolt.make_and_save_symbol,
     'desc': 'List symbols found in the goto binary',
     'flags': ['--viewer-symbol', '--viewer-source', '--goto', '--wkdir', '--srcdir']},
    {'name': 'trace',
     'func': tracet.make_and_save_trace,
     'desc': 'List error traces generated for issues found during property checking',
     'flags': ['--result', '--viewer-trace', '--wkdir', '--srcdir']},
]

def add_arguments(parser, function=None, flags=None):
    """Add implementing function and list of flags to a parser"""

    if function:
        parser.set_defaults(func=function)
    if flags:
        flags = flags + ['--verbose', '--debug', '--version']

    for group in OPTION_GROUPS:
        options = [
            option for option in group['group_opts'] if flags is None or option['flag'] in flags]
        if not options:
            continue

        parser_group = parser.add_argument_group(group['group_name'], group['group_desc'])
        for option in options:
            opt = dict(option) # pop is destructive
            flag = opt.pop('flag')
            parser_group.add_argument(flag, **opt)

def add_subparser(subparsers, sub_name, sub_desc, sub_func, sub_flags):
    """Add a defined subparser"""

    if sub_name:
        subparser = subparsers.add_parser(sub_name, description=sub_desc)
    add_arguments(subparser, sub_func, sub_flags)

def create_parser():
    """Create the parser with defined subparsers"""

    top = SUBPARSERS[0]
    assert top['name'] is None

    parser = argparse.ArgumentParser(description=top['desc'])
    add_arguments(parser, top['func'], top['flags'])

    subparsers = parser.add_subparsers(title='Subcommands', dest='subcommand')
    for subparser in SUBPARSERS[1:]:
        add_subparser(subparsers,
                      subparser['name'], subparser['desc'], subparser['func'], subparser['flags'])
    return parser

################################################################

def handle_deprecated_arguments(args):
    """Warn about deprecated arguments, use them  when possible."""

    if hasattr(args, 'block') and args.block:
        if hasattr(args, 'coverage'):
            logging.warning("--block is deprecated, using --coverage %s.",
                            args.block)
            args.coverage = [args.block] # block is a string, coverage is a list
        else:
            logging.warning("--block is deprecated, use --coverage instead.")
        args.block = None

    if hasattr(args, 'htmldir') and args.htmldir:
        if hasattr(args, 'reportdir'):
            logging.warning("--htmldir is deprecated, using --reportdir %s.",
                            args.htmldir)
            args.reportdir = args.htmldir
        else:
            logging.warning("--htmldir is deprecated, use --reportdir instead.")
        args.htmldir = None

    if hasattr(args, 'srcexclude') and args.srcexclude:
        if hasattr(args, 'exclude'):
            logging.warning("--srcexclude is deprecated, using --exclude %s.",
                            args.srcexclude)
            args.exclude = args.srcexclude
        else:
            logging.warning("--srcexclude is deprecated, "
                            "use --exclude instead.")
        logging.warning("--srcexclude and --exclude use slight different "
                        "regular expressions.")
        args.srcexclude = None

    if hasattr(args, 'blddir') and args.blddir:
        logging.warning("--blddir is deprecated, ignoring --blddir.")
        args.blddir = None

    if hasattr(args, 'storm') and args.storm:
        logging.warning("--storm is deprecated, ignoring --storm.")
        args.storm = None

    return args

################################################################
# Set default values for arguments

def default_logging(args):
    'Set default logging configuration.'

    # Only the first invocation of basicConfig configures the root logger
    if getattr(args, 'debug', False):
        logging.basicConfig(level=logging.DEBUG,
                            format='%(levelname)s: %(message)s')
    if getattr(args, 'verbose', False):
        logging.basicConfig(level=logging.INFO,
                            format='%(levelname)s: %(message)s')
    logging.basicConfig(format='%(levelname)s: %(message)s')

    return args

def default_source_method(args):
    'Set default source method.'

    # Set source_method to an enum
    if hasattr(args, 'source_method'):

        # Set source method to its enum value or None
        args.source_method = {
            'find': Sources.FIND,
            'walk': Sources.WALK,
            'make': Sources.MAKE,
            'goto': Sources.GOTO,
            None: None
        }[args.source_method]

        # Set source method to a reasonable default value
        if args.source_method is None:
            if getattr(args, 'srcdir') is not None:
                if getattr(args, 'wkdir') is not None:
                    if getattr(args, 'goto') is not None:
                        args.source_method = Sources.GOTO
                    else:
                        args.source_method = Sources.MAKE
                elif platform.system() == 'Windows':
                    args.source_method = Sources.WALK
                else:
                    args.source_method = Sources.FIND

        # Confirm existence of command line options needed by source method
        if args.source_method == Sources.GOTO:
            if (not hasattr(args, 'srcdir') or
                    not hasattr(args, 'wkdir') or
                    not hasattr(args, 'goto')):
                raise UserWarning('Options --srcdir, --wkdir, and --goto '
                                  'required by source method goto.')

        if args.source_method == Sources.FIND:
            if not hasattr(args, 'srcdir'):
                raise UserWarning('Option --srcdir required '
                                  'by source method find.')

        if args.source_method == Sources.WALK:
            if not hasattr(args, 'srcdir'):
                raise UserWarning('Option --srcdir required '
                                  'by source method walk.')

        if args.source_method == Sources.MAKE:
            if not hasattr(args, 'srcdir') or not hasattr(args, 'wkdir'):
                raise UserWarning('Options --srcdir and --wkdir required '
                                  'by source method make.')

    return args

def warn_against_using_text_for_cbmc_output(args):
    'Recommend the use of xml or json input instead of text.'

    for attr in ['result', 'coverage', 'property']:
        filenames = getattr(args, attr, None)
        if filenames and filet.any_text_files(filenames):
            logging.warning("Use xml or json instead of text for "
                            "better results: %s", filenames)

def defaults(args):
    'Set default values based on command line arguments.'

    args = default_logging(args)
    args = handle_deprecated_arguments(args)
    args = default_source_method(args)
    warn_against_using_text_for_cbmc_output(args)

    if hasattr(args, 'srcdir') and args.srcdir is not None:
        args.srcdir = os.path.abspath(args.srcdir)
    if hasattr(args, 'wkdir') and args.wkdir is not None:
        args.wkdir = os.path.abspath(args.wkdir)

    return args

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Annotated source code."""

import html
import logging
import os
import re

import voluptuous
import voluptuous.humanize

from cbmc_viewer import markup_link
from cbmc_viewer import templates
from cbmc_viewer import util

################################################################
# Data passed to jinja to generate annotated code from code.jinja.html

VALID_LINE = voluptuous.Schema({
    'num': int,
    'status': voluptuous.Any('hit', 'missed', 'both', 'none'),
    'code': str,
}, required=True)

VALID_CODE = voluptuous.Schema({
    'filename': str,
    'path_to_root': str,
    'lines': [VALID_LINE],
    'outdir': str
}, required=True)

################################################################
# An annotated file of source code.

class Code:
    """An annotated file of source code."""

    def __init__(self, root, path, symbols, coverage, outdir='.'):

        try:
            # load code into a string
            with open(os.path.join(root, path), encoding='utf-8') as source:
                code = html.escape(untabify_code(source.read()), quote=False)

            # split code into blocks of code, comments, and string literals
            blocks = split_code_into_blocks(code)

            # link symbols in code blocks to symbol definitions
            linked_blocks = link_symbols_in_code_blocks(path, blocks, symbols)

            # reform code as a string with symbols linked to definitions
            linked_code = ''.join(linked_blocks)

            # break code into lines annotated with line number and line coverage
            annotated_lines = annotate_code(path, linked_code, coverage)

            self.lines = annotated_lines
        except FileNotFoundError:
            # The goto symbol table occassional refers to header files
            # like gcc_builtin_headers_types.h that are part of the
            # CBMC implementation.
            #   * We skip source annotation: We treat the file as a
            #     zero-length file with nothing to annotate.
            #   * We print a simple warning message: The relative path
            #     to the file in the symbol table was interpreted by
            #     the symbol table parser as relative to the working
            #     directory.  Printing this path in the warning is
            #     confusing, so we print just the base name.
            logging.warning("Skipping source file annotation: %s",
                            os.path.basename(path))
            self.lines = []

        self.filename = path
        self.path_to_root = markup_link.path_to_file('.', path)
        self.outdir = outdir
        self.validate()

    def __str__(self):
        """Render annotated code as html."""

        return templates.render_code(self.filename,
                                     self.path_to_root,
                                     self.lines)

    def validate(self):
        """Validate members of a code object."""

        return voluptuous.humanize.validate_with_humanized_errors(
            self.__dict__, VALID_CODE
        )

    def dump(self, filename=None, outdir=None):
        """Write annotated code to a file rendered as html."""

        util.dump(self,
                  filename or self.filename + ".html",
                  outdir or self.outdir)

################################################################
# Untabify code: replace tabs with spaces.

def untabify_code(code, tabstop=8):
    """Untabify a block of code."""

    return '\n'.join(
        [untabify_line(line, tabstop) for line in code.splitlines()]
    )

def untabify_line(line, tabstop=8):
    """Untabify a line of code."""

    strings = []
    length = 0
    for string in re.split('(\t)', line):
        if string == '\t':
            string = ' '*(tabstop - (length % tabstop))
        length += len(string)
        strings.append(string)
    return ''.join(strings)

################################################################
# Split code into code blocks and strings/comments

def split_code_into_blocks(code):
    """Split code into blocks of code, comments, and string literals."""

    def is_noncode_start(code, idx=0):
        """String starts something other than source code (eg, a comment)."""

        return (is_quote(code, idx) or
                is_multiline_comment_start(code, idx) or
                is_singleline_comment_start(code, idx))

    def find_predicate(code, predicate, idx=0):
        """First position in a string satisfying a predicate."""

        while idx < len(code) and not predicate(code, idx):
            idx += 1
        return idx

    blocks = []

    while code:
        idx = find_predicate(code, is_noncode_start)
        block, code, idx = code[:idx], code[idx:], 0
        if block:
            blocks.append(block)

        if not code:
            break

        if is_quote(code):
            idx = find_predicate(code, is_quote, 1)
        elif is_multiline_comment_start(code):
            idx = find_predicate(code, is_multiline_comment_end, 2)
        elif is_singleline_comment_start(code):
            idx = find_predicate(code, is_singleline_comment_end, 2)
        block, code, idx = code[:idx+1], code[idx+1:], 0
        if block:
            blocks.append(block)

    return blocks

def is_quote(code, idx=0):
    """Position in string is an unescaped quotation mark."""
    return (0 <= idx < len(code) and
            code[idx] == '"' and
            (idx == 0 or code[idx-1] != '\\'))

def is_multiline_comment_start(code, idx=0):
    """Position in string starts a multi-line comment."""
    return idx >= 0 and idx+2 <= len(code) and code[idx:idx+2] == '/*'

def is_multiline_comment_end(code, idx=0):
    """Position in string ends a multi-line comment."""
    return idx-1 >= 0 and idx+1 <= len(code) and code[idx-1:idx+1] == '*/'

def is_singleline_comment_start(code, idx=0):
    """Position in string starts a one-line comment."""
    return idx >= 0 and idx+2 <= len(code) and code[idx:idx+2] == '//'

def is_singleline_comment_end(code, idx=0):
    """Position in string ends a one-line comment."""
    return idx >= 0 and idx+1 < len(code) and code[idx+1] == '\n'

################################################################
# Link symbols in code blocks

def link_symbols_in_code_blocks(path, blocks, symbols):
    """Link symbols appearing a sequence of blocks."""

    return [link_symbols_in_code_block(path, block, symbols)
            for block in blocks]

def link_symbols_in_code_block(path, block, symbols):
    """Link symbols appearing a code block."""

    if (is_quote(block) or
            is_multiline_comment_start(block) or
            is_singleline_comment_start(block)):
        return block

    return link_symbols(path, block, symbols)

def link_symbols(path, code, symbols):
    """Link symbols appearing a code string."""

    tokens = split_code_into_symbols(code)
    return ''.join(
        [markup_link.link_text_to_symbol(tkn, tkn, symbols, from_file=path, escape_text=False)
         for tkn in tokens]
    )

def split_code_into_symbols(code):
    """Split a code string into a list of symbols and nonsymbols."""

    return re.split('([_a-zA-Z][_a-zA-Z0-9]*)', code)

################################################################
# Annotate lines of symbol-linked code with line numbers and coverage

def annotate_code(path, code, coverage):
    """Annotate lines of code with line numbers and coverage status."""

    return [{ # line_num is 0-based, line numbers are 1-based
        'num': line_num+1,
        'status': str(coverage.lookup(path, line_num+1)).lower(),
        'code': line
    } for (line_num, line) in enumerate(code.splitlines())]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Symbol table parser for a goto binary symbol table."""

import logging
import os
import re

from cbmc_viewer import runt
from cbmc_viewer import srcloct

def symbol_table(goto):
    """Extract symbol table from goto binary as lines of text."""

    # The --show-symbol-table flag produces a sequence of symbol
    # definitions.  Definitions are separated by blank lines.  Each
    # definition is a sequence of lines including
    #
    #   Symbol......: symbol_name
    #   Pretty name.: simple_symbol_name
    #   Location....: file file_name line line_number

    # Code with definitions like
    #   static const char *UTF_16_BE_BOM = "\xFE\xFF";
    # will produce values in the symbol table that cannot be read as
    # strings of UTF-8 characters.  We read the symbol table using the
    # latin1 encoding in place of the Python default UTF-8 encoding,
    # since latin1 agrees with UTF-8 on the ASCII characters.

    cmd = ['cbmc', '--show-symbol-table', goto]
    definitions = re.split(r'[\n\r][\n\r]+', runt.run(cmd, encoding='latin1'))
    return [definition.strip().splitlines() for definition in definitions]

def is_symbol_line(line):
    """Line from symbol table defines a symbol name."""

    return line.startswith('Symbol.') # Match Symbol. not Symbols:

def is_location_line(line):
    """Line from symbol table defines a location of a symbol definition."""

    return line.startswith('Location')

def is_pretty_name_line(line):
    """Line from symbol table defines a simple symbol name (a pretty name)."""

    return line.startswith('Pretty name')

def parse_symbol(sym):
    """Symbol name from symbol line."""

    if not is_symbol_line(sym):
        return None

    # Examples of symbol lines
    # Symbol......: function_name
    # Symbol......: function_name$link1
    # Symbol......: function_name::parameter_name
    # Symbol......: function_name::1::1::variable_name
    # Symbol......: tag-struct_name
    # Symbol......: tag-union_name

    name = sym.split(":", 1)[1].strip()
    if name.startswith('tag-'):
        return name[len('tag-'):]
    if name:
        return name
    return None

def parse_location(loc, wkdir):
    """Symbol source location from location line."""

    if not is_location_line(loc):
        return None, None

    # Examples of location lines (may be no location for symbol)
    # Location....:
    # Location....: file file_name line line_number

    match = re.match('.* file (.*) line ([0-9]*)', loc)
    if match is None:
        return None, None

    rel_path, line = match.group(1), int(match.group(2))
    abs_path = srcloct.normpath(os.path.join(wkdir, rel_path))
    if srcloct.is_builtin(abs_path):
        return None, None

    return abs_path, line

def parse_pretty_name(sym):
    """Symbols pretty name from pretty name line."""

    if not is_pretty_name_line(sym):
        return None

    # Examples of pretty name lines (may be no pretty name for symbol)
    # Pretty name.:
    # Pretty name.: function_name
    # Pretty name.: function_name::1::1::variable_name
    # Pretty name.: struct struct_name
    # Pretty name.: union union_name

    name = sym.split(":", 1)[1].strip()
    if name.startswith('struct '):
        return name[len('struct '):]
    if name.startswith('union '):
        return name[len('union '):]
    if name:
        return name
    return None

def parse_symbol_table(definitions, wkdir):
    """Extract symbols and source locations from symbol table definitions."""

    def nonnull(items):
        nonnull_items = [item for item in items if item is not None]
        if nonnull_items:
            return nonnull_items[0]
        return None

    def nonnull_pair(items):
        nonnull_items = [item for item in items if item is not None
                         and (item[0] is not None or item[1] is not None)]
        if nonnull_items:
            return nonnull_items[0]
        return None, None

    def symbol(dfn):
        return nonnull([parse_symbol(line) for line in dfn])

    def pretty(dfn):
        return nonnull([parse_pretty_name(line) for line in dfn])

    def location(dfn, wkdir):
        return nonnull_pair([parse_location(line, wkdir) for line in dfn])

    def parse_definition(dfn, wkdir):
        loc = location(dfn, wkdir)
        return {
            'symbol': pretty(dfn) or symbol(dfn),
            'file': loc[0],
            'line': loc[1]
        }

    return [parse_definition(dfn, wkdir) for dfn in definitions]

def source_files(goto, wkdir, srcdir=None):
    """Source files appearing in symbol table.

    Source file path names in symbol table are absolute or relative to
    wkdir.  If srcdir is given, return only files under srcdir.
    """

    wkdir = srcloct.abspath(wkdir)
    srcs = [dfn['file']
            for dfn in parse_symbol_table(symbol_table(goto), wkdir)]
    srcs = [src for src in srcs if src and not srcloct.is_builtin(src)]

    if srcdir:
        srcdir = srcloct.abspath(srcdir)
        srcs = [src for src in srcs if src.startswith(srcdir)]

    return sorted(set(srcs))

def symbol_definitions(goto, wkdir, srcdir=None):
    """Symbol definitions appearing in symbol table.

    Source file path names in symbol table are absolute or relative to
    wkdir.  If srcdir is given, return only symbols defined in files
    under srcdir.
    """

    wkdir = srcloct.abspath(wkdir)
    srcdir = srcloct.abspath(srcdir)

    symbols = {}
    for dfn in parse_symbol_table(symbol_table(goto), wkdir):
        sym, src, num = dfn['symbol'], dfn['file'], dfn['line']

        if sym is None or src is None or num is None:
            logging.info("Skipping symbol table entry: %s: %s, %s",
                         sym, src, num)
            continue

        if srcdir and not src.startswith(srcdir):
            logging.info("Skipping symbol table entry: %s: %s, %s",
                         sym, src, num)
            continue

        srcloc = srcloct.make_srcloc(src, None, num, wkdir, srcdir)
        if sym in symbols and srcloc != symbols[sym]:
            logging.warning("Skipping redefinition of symbol name: %s", sym)
            logging.warning("  Old symbol %s: file %s, line %s",
                            sym, symbols[sym]["file"], symbols[sym]["line"])
            logging.warning("  New symbol %s: file %s, line %s",
                            sym, srcloc["file"], srcloc["line"])
            continue
        symbols[sym] = srcloc

    return symbols



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""CBMC property checking results."""

import json
import logging
import re

from voluptuous import Schema, Any
import voluptuous.humanize

from cbmc_viewer import filet
from cbmc_viewer import parse
from cbmc_viewer import propertyt
from cbmc_viewer import util
from cbmc_viewer.util import flatten, choose

JSON_TAG = 'viewer-result'

################################################################
# Validator for property checking results

PROGRAM = 'program'
STATUS = 'status'
WARNING = 'warning'
RESULT = 'results'
PROVER = 'prover'
FAILURE = 'failure'
SUCCESS = 'success'

VALID_RESULT = Schema({
    PROGRAM: Any(str, None), # program version string
    STATUS: [str],           # list of status messages
    WARNING: [str],          # list of warning messages
    RESULT: {
        True : [str],        # list of proved assertions
        False: [str]         # list of failed assertions
    },
    PROVER: Any(FAILURE, SUCCESS, None) # prover status: success or failure
}, required=True)

EMPTY_RESULT = {
    PROGRAM: None,
    STATUS: [],
    WARNING: [],
    RESULT: {
        True: [],
        False: [],
    },
    PROVER: None
}

################################################################
# CBMC property checking results

class Result:
    """CBMC property checking results."""

    def __init__(self, results_list=None):
        """Merge a list of results into a single result.

        Each result is given by the dict representation of a result object.
        """

        results_list = results_list or [EMPTY_RESULT]
        for results in results_list:
            self.validate(results)

        self.program = choose([results[PROGRAM] for results in results_list])
        self.status = flatten([results[STATUS] for results in results_list])
        self.warning = flatten([results[WARNING] for results in results_list])

        success = set(flatten([results[RESULT][True] for results in results_list]))
        failure = set(flatten([results[RESULT][False] for results in results_list]))
        self.results = {
            True: sorted(success.difference(failure), key=propertyt.key),
            False: sorted(failure, key=propertyt.key)
        }

        status = [results[PROVER] for results in results_list] or [None]
        self.prover = None if None in status else FAILURE if FAILURE in status else SUCCESS
        self.validate()

    def __repr__(self):
        """A dict representation of results."""

        self.validate()
        return self.__dict__

    def __str__(self):
        """A string representation of results."""

        return json.dumps({JSON_TAG: self.__repr__()}, indent=2, sort_keys=True)

    def validate(self, results=None):
        """Validate results."""

        return voluptuous.humanize.validate_with_humanized_errors(
            results or self.__dict__, VALID_RESULT
        )

    def dump(self, filename=None, directory=None):
        """Write results to a file or stdout."""

        util.dump(self, filename, directory)

################################################################
# CBMC property checking results loaded from make-result json data

class ResultFromJson(Result):
    """CBMC property checking results loaded from make-result json data"""

    def __init__(self, json_files):
        super().__init__(
            [self.load_results(json_file) for json_file in json_files]
        )

    @staticmethod
    def load_results(json_file):
        """Load CBMC results from json file"""

        blob = parse.parse_json_file(json_file)[JSON_TAG]

        # json uses strings "true" and "false" for True and False keys
        blob[RESULT][True] = blob[RESULT]["true"]
        blob[RESULT][False] = blob[RESULT]["false"]
        del blob[RESULT]["true"]
        del blob[RESULT]["false"]

        return blob

################################################################
# CBMC property checking results loaded from cbmc text output

class ResultFromCbmcText(Result):
    """CBMC property checking results loaded from cbmc text output"""

    def __init__(self, text_files):
        super().__init__(
            [parse_cbmc_text_results(text_file) for text_file in text_files]
        )

################################################################
# CBMC property checking results loaded from cbmc json output

class ResultFromCbmcJson(Result):
    """CBMC property checking results loaded from cbmc json output"""

    def __init__(self, json_files):
        super().__init__(
            [parse_cbmc_json_results(json_file) for json_file in json_files]
        )

################################################################
# CBMC property checking results loaded from cbmc xml output

class ResultFromCbmcXml(Result):
    """CBMC property checking results loaded from CBMC xml output"""

    def __init__(self, xml_files):
        super().__init__(
            [parse_cbmc_xml_results(xml_file) for xml_file in xml_files]
        )

################################################################
# Parse the text output of cbmc property checking into sections
#
# This parsing is required only for text output of cbmc.  The text
# output is just an unstructured sequence of lines with embedded
# section headers.  The json and xml output of cbmc are already
# structured into sections, and sections are easy to find.

LOG_SECTION = 'log'
RESULTS_SECTION = 'results'
TRACES_SECTION = 'traces'
SUMMARY_SECTION = 'summary'

EMPTY_SECTIONS = {
    LOG_SECTION: [],
    RESULTS_SECTION: [],
    TRACES_SECTION: [],
    SUMMARY_SECTION: []
}

RESULTS_SECTION_HEADER = "** Results"
TRACES_SECTION_HEADER = "Trace for"
SUMMARY_SECTION_HEADER = "** " # Caution: summary header is a prefix of results header

def cbmc_text_sections(text_file):
    """Parse the text output of cbmc property checking into sections"""

    sections = EMPTY_SECTIONS
    with open(text_file, encoding='utf-8') as blob:
        # The blob is an iterator that iterates through the lines of
        # the file and throws StopIteration at the end of the file.
        next_line = lambda: next(blob).rstrip()
        try:
            line = next_line()

            # Log section is from here to results section
            while not line.startswith(RESULTS_SECTION_HEADER):
                sections[LOG_SECTION].append(line)
                line = next_line()

            # Results section is from here to traces section or summary section
            # Traces section is optional: generated by `cbmc --trace`
            line = next_line() # skip results header
            while (not line.startswith(TRACES_SECTION_HEADER) and
                   not line.startswith(SUMMARY_SECTION_HEADER)):
                sections[RESULTS_SECTION].append(line)
                line = next_line()

            # Traces section is from here to summary section
            # Traces section is optional: generated by `cbmc --trace`
            if line.startswith(TRACES_SECTION_HEADER):
                while not line.startswith(SUMMARY_SECTION_HEADER):
                    sections[TRACES_SECTION].append(line)
                    line = next_line()

            # Summary section is from here to end of file
            while True:
                sections[SUMMARY_SECTION].append(line)
                line = next_line()
        except StopIteration:
            return sections

################################################################
# Parse cbmc property checking output

def parse_cbmc_text_results(text_file):
    """Parse text output of cbmc property checking"""

    try:
        sections = cbmc_text_sections(text_file)
    except OSError as error:
        raise UserWarning(f"Can't parse {text_file}: {str(error)}") from error

    results = dict(EMPTY_RESULT)
    results[PROGRAM] = cbmc_text_program(sections[LOG_SECTION])
    results[STATUS] = cbmc_text_status(sections[LOG_SECTION], sections[SUMMARY_SECTION])
    results[WARNING] = cbmc_text_warnings(sections[LOG_SECTION])
    results[RESULT] = cbmc_text_results(sections[RESULTS_SECTION])
    results[PROVER] = cbmc_text_prover(sections[SUMMARY_SECTION])

    return results

def parse_cbmc_json_results(json_file):
    """Parse json output of cbmc property checking"""

    blob = parse.parse_json_file(json_file)
    if blob is None:
        return EMPTY_RESULT

    results = dict(EMPTY_RESULT)
    results[PROGRAM] = cbmc_json_program(blob)
    results[STATUS] = cbmc_json_status(blob)
    results[WARNING] = cbmc_json_warnings(blob)
    results[RESULT] = cbmc_json_results(blob)
    results[PROVER] = cbmc_json_prover(blob)

    return results

def parse_cbmc_xml_results(xml_file):
    """Parse xml output of cbmc property checking"""

    blob = parse.parse_xml_file(xml_file)
    if blob is None:
        return EMPTY_RESULT

    results = dict(EMPTY_RESULT)
    results[PROGRAM] = cbmc_xml_program(blob)
    results[STATUS] = cbmc_xml_status(blob)
    results[WARNING] = cbmc_xml_warnings(blob)
    results[RESULT] = cbmc_xml_results(blob)
    results[PROVER] = cbmc_xml_prover(blob)

    return results

# Keys used in cbmc json output

JSON_DESCRIPTION_KEY = 'description'
JSON_MESSAGE_TEXT_KEY = 'messageText'
JSON_MESSAGE_TYPE_KEY = 'messageType'
JSON_PROGRAM_KEY = 'program'
JSON_PROPERTY_KEY = 'property'
JSON_PROVER_STATUS_KEY = 'cProverStatus'
JSON_RESULT_KEY = 'result'
JSON_STATUS_KEY = 'status'

JSON_STATUS_MESSAGE = 'STATUS-MESSAGE'
JSON_SUCCESS = 'SUCCESS'
JSON_WARNING = 'WARNING'

JSON_PROVER_STATUS_SUCCESS = 'success'

# Keys used in cbmc xml output

XML_FAILURE_TAG = 'failure'
XML_GOTO_TRACE_TAG = 'goto_trace'
XML_LOCATION_TAG = 'location'
XML_MESSAGE_TAG = 'message'
XML_MESSAGE_TEXT_TAG = 'text'
XML_PROGRAM_TAG = 'program'
XML_PROVER_STATUS_TAG = 'cprover-status'
XML_RESULT_TAG = 'result'

XML_FAILURE_PROPERTY_ATTR = 'property'
XML_FAILURE_REASON_ATTR = 'reason'
XML_MESSAGE_TYPE_ATTR = 'type'
XML_RESULT_PROPERTY_ATTR = 'property'
XML_RESULT_STATUS_ATTR = 'status'

XML_STATUS_MESSAGE = 'STATUS-MESSAGE'
XML_SUCCESS_STATUS = 'SUCCESS'
XML_WARNING_MESSAGE = 'WARNING'

################################################################
# Find PROGRAM in cbmc property checking output

def cbmc_text_program(log_section):
    """Find program in cbmc text output"""

    for line in log_section:
        if line.startswith('CBMC version'):
            return line
    return None

def cbmc_json_program(blobs):
    """Find program in cbmc json output"""

    for blob in blobs:
        program = blob.get(JSON_PROGRAM_KEY)
        if program:
            return program
    return None

def cbmc_xml_program(blob):
    """Find program in cbmc xml output"""

    program = blob.find(XML_PROGRAM_TAG)
    if program is not None:
        return program.text
    return None

################################################################
# Find STATUS messages in cbmc property checking output

def cbmc_text_status(log_section, summary_section):
    """Find status messages in cbmc text output"""

    status = []
    for line in log_section:
        if line and not line.startswith('**** WARNING:'):
            status.append(line)
    for line in summary_section:
        if line and line.startswith('VERIFICATION '):
            status.append(line)
    return status

def cbmc_json_status(blobs):
    """Find status messages in cbmc json output"""

    status = []
    for blob in blobs:
        kind = blob.get(JSON_MESSAGE_TYPE_KEY)
        text = blob.get(JSON_MESSAGE_TEXT_KEY)
        if kind == JSON_STATUS_MESSAGE:
            status.append(text)
    return status

def cbmc_xml_status(blob):
    """Find status messages in cbmc xml output"""

    status = []
    for msg in blob.iter(XML_MESSAGE_TAG):
        if msg.get(XML_MESSAGE_TYPE_ATTR) == XML_STATUS_MESSAGE:
            status.append(msg.find(XML_MESSAGE_TEXT_TAG).text)
    return status

################################################################
# Find WARNING messages in cbmc property checking output

def cbmc_text_warnings(log_section):
    """Find warning messages in cbmc text output"""

    warnings = []
    for line in log_section:
        if line.startswith('**** WARNING:'):
            warnings.append(line)
    return warnings

def cbmc_json_warnings(blobs):
    """Find warning messages in cbmc json output"""

    warnings = []
    for blob in blobs:
        kind = blob.get(JSON_MESSAGE_TYPE_KEY)
        text = blob.get(JSON_MESSAGE_TEXT_KEY)
        if kind == JSON_WARNING:
            warnings.append(text)
    return warnings

def cbmc_xml_warnings(blob):
    """Find warning messages in cbmc xml output"""

    warnings = []
    for msg in blob.iter(XML_MESSAGE_TAG):
        if msg.get(XML_MESSAGE_TYPE_ATTR) == XML_WARNING_MESSAGE:
            warnings.append(msg.find(XML_MESSAGE_TEXT_TAG).text)
    return warnings

################################################################
# Find RESULTS in cbmc property checking output

EMPTY_RESULT_RESULTS = {True: [], False: []}

def cbmc_text_results(results_section):
    """Find results in cbmc text output"""

    results = EMPTY_RESULT_RESULTS
    for line in results_section:
        # Lines given property checking results have the form
        # [name] srcloc description: SUCCESS|FAILURE
        match = re.match(r'\[([^ ]*)\].*: ((FAILURE)|(SUCCESS))', line)
        if match:
            name, status = match.groups()[:2]
            results[status == 'SUCCESS'].append(name)
    return results

def cbmc_json_results(blobs):
    """Find results in cbmc json output"""

    results = EMPTY_RESULT_RESULTS
    for blob in blobs:
        for result in blob.get(JSON_RESULT_KEY) or []:
            name, status = result[JSON_PROPERTY_KEY], result[JSON_STATUS_KEY]
            results[status == JSON_SUCCESS].append(name)
    return results

def cbmc_xml_results(blob):
    """Find results in cbmc xml output"""

    results = EMPTY_RESULT_RESULTS

    # cbmc normal output produces property checking results for all properties
    for result in blob.iter(XML_RESULT_TAG):
        name = result.get(XML_RESULT_PROPERTY_ATTR)
        status = result.get(XML_RESULT_STATUS_ATTR)
        results[status == XML_SUCCESS_STATUS].append(name)

    # cbmc --stop-on-fail output produces one trace with one failure
    if not (results[True] or results[False]):
        trace = blob.find(XML_GOTO_TRACE_TAG)
        if trace is not None:
            results[False] = [trace.find(XML_FAILURE_TAG).get(XML_FAILURE_PROPERTY_ATTR)]

    return results

################################################################
# Find PROVER status in cbmc property checking output

def cbmc_text_prover(summary_section):
    """Find prover status in cbmc text output"""

    # Infer prover status from the lines 'VERIFICATION SUCCESSFUL' or
    # 'VERIFICATION FAILED' in the summary section.

    # In rare cases, cbmc can solve the constraint problem without
    # invoking the constraint solver, and the string '** X of Y
    # failed` that we use as the summary header is not printed. The
    # string 'VERIFICATION SUCCESSFUL' that we look for in the summary
    # section is absorbed into the log section, and the summary
    # section is empty.

    if not summary_section:
        return SUCCESS

    for line in summary_section:
        if line == 'VERIFICATION FAILED':
            return FAILURE
        if line == 'VERIFICATION SUCCESSFUL':
            return SUCCESS
    return None

def cbmc_json_prover(blobs):
    """Find prover status in cbmc json output"""

    for blob in blobs:
        prover = blob.get(JSON_PROVER_STATUS_KEY)
        if prover:
            return SUCCESS if prover == JSON_PROVER_STATUS_SUCCESS else FAILURE
    return None

def cbmc_xml_prover(blob):
    """Find prover status in cbmc xml output"""

    prover = blob.find(XML_PROVER_STATUS_TAG)
    if prover is not None:
        return SUCCESS if prover.text == XML_SUCCESS_STATUS else FAILURE
    return None

################################################################
# make-result

def fail(msg):
    """Log failure and raise exception."""

    logging.info(msg)
    raise UserWarning(msg)

def make_result(args):
    """Implementation of make-result."""

    viewer_result, cbmc_result = args.viewer_result, args.result

    if viewer_result:
        if filet.all_json_files(viewer_result):
            return ResultFromJson(viewer_result)
        fail(f"Expected a list of json files: {viewer_result}")

    if cbmc_result:
        if filet.all_text_files(cbmc_result):
            return ResultFromCbmcText(cbmc_result)
        if filet.all_json_files(cbmc_result):
            return ResultFromCbmcJson(cbmc_result)
        if filet.all_xml_files(cbmc_result):
            return ResultFromCbmcXml(cbmc_result)
        fail(f"Expected a list of text files, json files, or xml files: {cbmc_result}")

    logging.info("make-result: nothing to do: need "
                 "cbmc property checking results or "
                 "--viewer-result")
    return Result()

def make_and_save_result(args, path=None):
    """Make result object and write to file or stdout"""

    obj = make_result(args)
    util.save(obj, path)
    return obj

################################################################



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Jinja templates."""

import jinja2
from jinja2 import select_autoescape

import pkg_resources

PACKAGE = 'cbmc_viewer'
TEMPLATES = 'templates'

SUMMARY_TEMPLATE = 'summary.jinja.html'
CODE_TEMPLATE = 'code.jinja.html'
TRACE_TEMPLATE = 'trace.jinja.html'

ENV = None

def env():
    """The jinja environment."""

    # pylint: disable=global-statement
    global ENV

    if ENV is None:
        template_dir = pkg_resources.resource_filename(PACKAGE, TEMPLATES)
        ENV = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir),
            autoescape=select_autoescape(
                enabled_extensions=('html'),
                default_for_string=True)
        )
    return ENV

def render_summary(summary):
    """Render summary as html."""

    return env().get_template(SUMMARY_TEMPLATE).render(
        summary=summary
    )

def render_code(filename, path_to_root, lines):
    """Render annotated source code as html."""

    return env().get_template(CODE_TEMPLATE).render(
        filename=filename, path_to_root=path_to_root, lines=lines
    )

def render_trace(name, desc, srcloc, steps):
    """Render annotated trace as html."""

    return env().get_template(TRACE_TEMPLATE).render(
        prop_name=name, prop_desc=desc, prop_srcloc=srcloc, steps=steps
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""Version number."""

NAME = "CBMC viewer"
NUMBER = "3.3"
VERSION = f"{NAME} {NUMBER}"

def version(display=False):
    """The version of cbmc viewer."""

    if display:
        print(VERSION)
    return VERSION



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""The known file types."""

import enum
import os

################################################################

class File(enum.Enum):
    """The known file types."""

    TEXT = 1
    XML = 2
    JSON = 3

def filetype(filename):
    """Return the file type denoted by the filename extension."""

    if filename is None:
        return None

    if not isinstance(filename, str):
        raise UserWarning(f"Filename is not a string: {filename}")

    # The filename extension: expected to be one of txt, jsn, json, or xml
    file_extension = os.path.splitext(filename)[1].lower().lstrip('.')

    # Return the file type
    try:
        return {
            'log': File.TEXT,
            'txt': File.TEXT,
            'jsn': File.JSON,
            'json': File.JSON,
            'xml': File.XML
        }[file_extension]
    except KeyError:
        raise UserWarning(
            f"Can't determine file type of file {filename}"
        ) from None # squash the KeyError context, raise just a UserWarning

################################################################

def is_text_file(txt):
    """File is a text file."""
    return filetype(txt) == File.TEXT
def is_json_file(json):
    """File is a json file."""
    return filetype(json) == File.JSON
def is_xml_file(xml):
    """File is an xml file."""
    return filetype(xml) == File.XML

def all_text_files(txts):
    """Files are text files."""
    return all(is_text_file(txt) for txt in txts)
def all_json_files(jsons):
    """Files are json files."""
    return all(is_json_file(json) for json in jsons)
def all_xml_files(xmls):
    """Files are xml files."""
    return all(is_xml_file(xml) for xml in xmls)

def any_text_files(txts):
    """Any of the files are text files."""
    return any(is_text_file(txt) for txt in txts)

################################################################
